#!/usr/bin/env python
"""
Map panel of ncvue.

The panel allows plotting contour or mesh maps of georeferenced data.
Maps can be animated along the time axis.

This module was written by Matthias Cuntz while at Institut National de
Recherche pour l'Agriculture, l'Alimentation et l'Environnement (INRAE), Nancy,
France.

:copyright: Copyright 2020-2021 Matthias Cuntz - mc (at) macu (dot) de
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following classes are provided:

.. autosummary::
   ncvMap

History
   * Written Dec 2020-Jan 2021 by Matthias Cuntz (mc (at) macu (dot) de)
   * Open new netcdf file, communicate via top widget,
     Jan 2021, Matthias Cuntz
   * Write coordinates and value on bottom of plotting canvas,
     May 2021, Matthias Cuntz
   * Larger pad for colorbar, Jun 2021, Matthias Cuntz
   * Work with files without an unlimited (time) dimension (set_tstep),
     Oct 2021, Matthias Cuntz
   * Address fi.variables[name] directly by fi[name], Jan 2024, Matthias Cuntz
   * Allow groups in netcdf files, Jan 2024, Matthias Cuntz
   * Allow multiple netcdf files, Jan 2024, Matthias Cuntz
   * Move images/ directory from src/ncvue/ to src/ directory,
     Jan 2024, Matthias Cuntz
   * Added borders, rivers, and lakes checkbuttons, Feb 2024, Matthias Cuntz
   * Move themes/ and images/ back to src/ncvue/, Feb 2024, Matthias Cuntz
   * Use matplotlib.colormaps[name] instead of
     matplotlib.colormaps.get_cmap(name), Jul 2024, Matthias Cuntz
   * Use draw_idle instead of draw in update method for faster animation,
     Jul 2024, Matthias Cuntz
   * Add Quit button, Nov 2024, Matthias Cuntz
   * Use CustomTkinter if installed, Nov 2024, Matthias Cuntz
   * Use add_button, add_label widgets, Feb 2025, Matthias Cuntz
   * Use add_combobox instead of Combobox directly, Feb 2025, Matthias Cuntz
   * Remove delay, Feb 2025, Matthias Cuntz
   * Include xarray to read input files, Feb 2025, Matthias Cuntz

"""
import os
import sys
import tkinter as tk
try:
    from customtkinter import CTkFrame as Frame
    ihavectk = True
except ModuleNotFoundError:
    from tkinter.ttk import Frame
    ihavectk = False
import cartopy.crs as ccrs
import cartopy.feature as cfeature
import netCDF4 as nc
import numpy as np
try:
    import xarray as xr
    ihavex = True
except ModuleNotFoundError:
    ihavex = False
from .ncvutils import add_cyclic, clone_ncvmain, format_coord_map, selvar
from .ncvutils import set_axis_label, set_miss, vardim2var
from .ncvmethods import analyse_netcdf, get_slice_miss, get_miss
from .ncvmethods import set_dim_lon, set_dim_lat, set_dim_var
from .ncvwidgets import add_button, add_checkbutton, add_combobox, add_entry
from .ncvwidgets import add_imagemenu, add_label, add_menu, add_scale
from .ncvwidgets import add_spinbox
import matplotlib as mpl
from matplotlib import pyplot as plt
try:
    # plt.style.use('seaborn-v0_8-darkgrid')
    plt.style.use('seaborn-v0_8-dark')
except OSError:
    # plt.style.use('seaborn-darkgrid')
    plt.style.use('seaborn-dark')
# plt.style.use('fast')


__all__ = ['ncvMap']


class ncvMap(Frame):
    """
    Panel for maps.

    Sets up the layout with the figure canvas, variable selectors, dimension
    spinboxes, and options in __init__.

    Contains various commands that manage what will be drawn or redrawn if
    something is selected, changed, checked, etc.

    """

    #
    # Panel setup
    #

    def __init__(self, master, **kwargs):
        from functools import partial
        from matplotlib import animation
        from matplotlib.figure import Figure
        from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
        from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk

        super().__init__(master, **kwargs)

        self.name   = 'Map'
        self.master = master
        self.top    = master.top
        # copy for ease of use
        self.usex   = self.top.usex
        self.fi     = self.top.fi
        self.groups = self.top.groups
        self.miss   = self.top.miss
        self.dunlim = self.top.dunlim
        self.time   = self.top.time
        self.tname  = self.top.tname
        self.tvar   = self.top.tvar
        self.dtime  = self.top.dtime
        self.latvar = self.top.latvar
        self.lonvar = self.top.lonvar
        self.latdim = self.top.latdim
        self.londim = self.top.londim
        self.maxdim = self.top.maxdim
        self.cols   = self.top.cols

        # unlimited dimension control
        self.iunlim = -1  # index of dunlim in dimensions of current var
        self.nunlim = 0   # length of dunlim of current plot variable

        # selections and options
        columns = [''] + self.cols

        allcmaps = plt.colormaps()
        self.cmaps  = [ i for i in allcmaps if not i.endswith('_r') ]
        self.cmaps.sort()
        bundle_dir = getattr(sys, '_MEIPASS',
                             os.path.abspath(os.path.dirname(__file__)))
        self.imaps  = [ tk.PhotoImage(file=f'{bundle_dir}/images/{i}.png')
                        for i in self.cmaps ]

        # only projections with keyword: central_longitude
        self.projs = ['AlbersEqualArea', 'AzimuthalEquidistant', 'EckertI',
                      'EckertII', 'EckertIII', 'EckertIV', 'EckertV',
                      'EckertVI', 'EqualEarth', 'EquidistantConic',
                      'InterruptedGoodeHomolosine',
                      'LambertAzimuthalEqualArea', 'LambertConformal',
                      'LambertCylindrical', 'Mercator', 'Miller', 'Mollweide',
                      'NorthPolarStereo', 'PlateCarree', 'Robinson',
                      'Sinusoidal', 'SouthPolarStereo', 'Stereographic']
        self.iprojs = [ccrs.AlbersEqualArea, ccrs.AzimuthalEquidistant,
                       ccrs.EckertI, ccrs.EckertII, ccrs.EckertIII,
                       ccrs.EckertIV, ccrs.EckertV, ccrs.EckertVI,
                       ccrs.EqualEarth, ccrs.EquidistantConic,
                       ccrs.InterruptedGoodeHomolosine,
                       ccrs.LambertAzimuthalEqualArea, ccrs.LambertConformal,
                       ccrs.LambertCylindrical, ccrs.Mercator, ccrs.Miller,
                       ccrs.Mollweide, ccrs.NorthPolarStereo, ccrs.PlateCarree,
                       ccrs.Robinson, ccrs.Sinusoidal, ccrs.SouthPolarStereo,
                       ccrs.Stereographic]
        if ihavectk:
            # width of combo boxes in px
            combowidth = 297
            smallcombowidth = 81
            # widths of entry widgets in px
            ewsmall = 18
            ewmed = 45
            ewbig = 72
            ewvbig = 117
            # pad between label and entry
            padx = 5
            # width of animation and variables buttons
            bwidth = 35
            # width of projections menu
            mwidth = 70
        else:
            # width of combo boxes in characters
            combowidth = 32
            smallcombowidth = 5
            # widths of entry widgets in characters
            ewsmall = 3
            ewmed = 4
            ewbig = 7
            ewvbig = 11
            # pad between label and entry (not used)
            padx = 5
            # width of animation and variables buttons
            bwidth = 2
            # width of projections menu
            mwidth = 13

        # open file and new window buttons
        self.rowwin = Frame(self)
        self.rowwin.pack(side=tk.TOP, fill=tk.X)
        self.newfile, self.newfiletip = add_button(
            self.rowwin, text='Open File', command=self.newnetcdf,
            tooltip='Open new netcdf file(s)')
        if ihavex:
            self.newxarray, self.newxarraytip = add_button(
                self.rowwin, text='Open xarray', command=self.newxarray,
                tooltip='Open new netcdf file(s) with xarray')
        spacew = add_label(self.rowwin, text='   ')
        time_label1 = add_label(self.rowwin, text='Time: ')
        self.timelbl, time_label2 = add_label(self.rowwin, '')
        self.newwin, self.newwintip = add_button(
            self.rowwin, text='New Window', nopack=True,
            command=partial(clone_ncvmain, self.master),
            tooltip='Open secondary ncvue window')
        self.newwin.pack(side=tk.RIGHT)

        # plotting canvas
        self.figure = Figure(facecolor='white', figsize=(1, 1))
        self.axes = self.figure.add_subplot(
            111, projection=ccrs.PlateCarree())
        self.canvas = FigureCanvasTkAgg(self.figure, master=self)
        self.canvas.draw()
        self.canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=1)

        # matplotlib toolbar
        self.toolbar = NavigationToolbar2Tk(self.canvas, self)
        self.toolbar.update()
        self.toolbar.pack(side=tk.TOP, fill=tk.X)

        # 1. row
        # controls
        self.rowt = Frame(self)
        self.rowt.pack(side=tk.TOP, fill=tk.X)
        ntime = 1
        self.tstepframe, self.tsteplbl, self.tstepval, self.tstep, self.tsteptip = (
            add_scale(self.rowt, label='step', ini=0, from_=0, to=ntime,
                      length=100, orient=tk.HORIZONTAL, command=self.tstep_t,
                      tooltip='Slide to go to time step'))
        self.tstepframe.pack(side=tk.LEFT)
        spacet = add_label(self.rowt, text=' ' * 1)

        # first t, previous t, etc.
        self.first_time, self.first_timetip = add_button(
            self.rowt, text='|<<', command=self.first_t, width=bwidth,
            tooltip='First time step')
        self.prev_time, self.prev_timetip = add_button(
            self.rowt, text='|<', command=self.prev_t, width=bwidth,
            tooltip='Previous time step')
        self.prun_timelbl, self.prun_time, self.prun_timetip = add_button(
            self.rowt, '<', command=self.prun_t, width=bwidth,
            tooltip='Run backwards')
        self.nrun_timelbl, self.nrun_time, self.nrun_timetip = add_button(
            self.rowt, '>', command=self.nrun_t, width=bwidth,
            tooltip='Run forwards')
        self.next_time, self.next_timetip = add_button(
            self.rowt, text='>|', command=self.next_t, width=bwidth,
            tooltip='Next time step')
        self.last_time, self.last_timetip = add_button(
            self.rowt, text='>>|', command=self.last_t, width=bwidth,
            tooltip='Last time step')
        # repeat
        spacer = add_label(self.rowt, text=' ' * 1)
        reps = ['once', 'repeat', 'reflect']
        tstr = ('Run time steps once, repeat from start when at end,'
                ' or continue running backwards when at end')
        self.repeatframe, self.repeatlbl, self.repeat, self.repeattip = (
            add_combobox(self.rowt, label='repeat', values=reps,
                         width=smallcombowidth, padx=padx,
                         command=self.repeat_t, tooltip=tstr))
        self.repeat.set('repeat')
        self.repeatframe.pack(side=tk.LEFT)

        # 2. row
        # variable-axis selection
        self.rowvv = Frame(self)
        self.rowvv.pack(side=tk.TOP, fill=tk.X)

        self.blockv = Frame(self.rowvv)
        self.blockv.pack(side=tk.LEFT)

        self.rowv = Frame(self.blockv)
        self.rowv.pack(side=tk.TOP, fill=tk.X)
        lkwargs = {}
        if ihavectk:
            lkwargs.update({'padx': padx})
        vlab = add_label(self.rowv, text='var', **lkwargs)
        spacep = add_label(self.rowv, text=' ' * 1)
        self.bprev_v, self.bprev_vtip = add_button(
            self.rowv, text='<', command=self.prev_v, width=bwidth,
            tooltip='Previous variable')
        self.bnext_v, self.bnext_vtip = add_button(
            self.rowv, text='>', command=self.next_v, width=bwidth,
            tooltip='Next variable')
        self.vframe, self.vlbl, self.v, self.vtip = add_combobox(
            self.rowv, label='', values=columns,
            command=self.selected_v, width=combowidth, padx=0,
            tooltip='Choose variable')
        self.vframe.pack(side=tk.LEFT)
        self.trans_vframe, self.trans_vlbl, self.trans_v, self.trans_vtip = (
            add_checkbutton(
                self.rowv, label='transpose var', value=False,
                command=self.checked,
                tooltip='Transpose array, i.e. exchanging lat and lon'))
        self.trans_vframe.pack(side=tk.LEFT)
        spacev = add_label(self.rowv, text=' ' * 1)
        self.vminframe, self.vminlbl, self.vmin, self.vmintip = add_entry(
            self.rowv, label='vmin', text=0, width=ewvbig,
            command=self.entered_v,
            tooltip='Minimal display value', padx=padx)
        self.vminframe.pack(side=tk.LEFT)
        self.vmaxframe, self.vmaxlbl, self.vmax, self.vmaxtip = add_entry(
            self.rowv, label='vmax', text=1, width=ewvbig,
            command=self.entered_v,
            tooltip='Maximal display value', padx=padx)
        self.vmaxframe.pack(side=tk.LEFT)
        tstr  = ('If checked, determine vmin/vmax from all fields,\n'
                 'otherwise from 50 random fields')
        self.vallframe, self.valllbl, self.vall, self.valltip = (
            add_checkbutton(self.rowv, label='all', value=False,
                            command=self.checked_all, tooltip=tstr))
        self.vallframe.pack(side=tk.LEFT)

        # levels var
        self.rowvd = Frame(self.blockv)
        self.rowvd.pack(side=tk.TOP, fill=tk.X)
        self.vdframe  = []
        self.vdlblval = []
        self.vdlbl    = []
        self.vdval    = []
        self.vd       = []
        self.vdtip    = []
        for i in range(self.maxdim):
            vdframe, vdlblval, vdlbl, vdval, vd, vdtip = add_spinbox(
                self.rowvd, label=str(i), values=(0,), wrap=True,
                command=self.spinned_v, state=tk.DISABLED, tooltip='None')
            self.vdframe.append(vdframe)
            self.vdlblval.append(vdlblval)
            self.vdlbl.append(vdlbl)
            self.vdval.append(vdval)
            self.vd.append(vd)
            self.vdtip.append(vdtip)
            vdframe.pack(side=tk.LEFT)

        # 3. row
        # lon-axis selection
        self.rowll = Frame(self)
        self.rowll.pack(side=tk.TOP, fill=tk.X)

        self.blocklon = Frame(self.rowll)
        self.blocklon.pack(side=tk.LEFT)

        self.rowlon = Frame(self.blocklon)
        self.rowlon.pack(side=tk.TOP, fill=tk.X)
        self.lonframe, self.lonlbl, self.lon, self.lontip = add_combobox(
            self.rowlon, label='lon', values=columns, padx=padx,
            command=self.selected_lon, width=combowidth,
            tooltip='Longitude variable.\nSet "empty" for matrix plot.')
        self.lonframe.pack(side=tk.LEFT)
        self.inv_lonframe, self.inv_lonlbl, self.inv_lon, self.inv_lontip = (
            add_checkbutton(self.rowlon, label='invert lon', value=False,
                            command=self.checked,
                            tooltip='Invert longitudes'))
        self.inv_lonframe.pack(side=tk.LEFT)
        self.shift_lonframe, self.shift_lonlbl, self.shift_lon, self.shift_lontip = (
            add_checkbutton(self.rowlon, label='shift lon/2', value=False,
                            command=self.checked,
                            tooltip='Roll longitudes by half its size'))
        self.shift_lonframe.pack(side=tk.LEFT)

        self.rowlond = Frame(self.blocklon)
        self.rowlond.pack(side=tk.TOP, fill=tk.X)
        self.londframe  = []
        self.londlblval = []
        self.londlbl    = []
        self.londval    = []
        self.lond       = []
        self.londtip    = []
        for i in range(self.maxdim):
            londframe, londlblval, londlbl, londval, lond, londtip = (
                add_spinbox(self.rowlond, label=str(i), values=(0,), wrap=True,
                            command=self.spinned_lon, state=tk.DISABLED,
                            tooltip='None'))
            self.londframe.append(londframe)
            self.londlblval.append(londlblval)
            self.londlbl.append(londlbl)
            self.londval.append(londval)
            self.lond.append(lond)
            self.londtip.append(londtip)
            londframe.pack(side=tk.LEFT)

        # lat-axis selection
        spacex = add_label(self.rowll, text=' ' * 3)

        self.blocklat = Frame(self.rowll)
        self.blocklat.pack(side=tk.LEFT)

        self.rowlat = Frame(self.blocklat)
        self.rowlat.pack(side=tk.TOP, fill=tk.X)
        self.latframe, self.latlbl, self.lat, self.lattip = add_combobox(
            self.rowlat, label='lat', values=columns, padx=padx,
            command=self.selected_lat, width=combowidth,
            tooltip='Longitude variable.\nSet "empty" for matrix plot.')
        self.latframe.pack(side=tk.LEFT)
        self.inv_latframe, self.inv_latlbl, self.inv_lat, self.inv_lattip = (
            add_checkbutton(self.rowlat, label='invert lat', value=False,
                            command=self.checked,
                            tooltip='Invert longitudes'))
        self.inv_latframe.pack(side=tk.LEFT)

        self.rowlatd = Frame(self.blocklat)
        self.rowlatd.pack(side=tk.TOP, fill=tk.X)
        self.latdframe  = []
        self.latdlblval = []
        self.latdlbl    = []
        self.latdval    = []
        self.latd       = []
        self.latdtip    = []
        for i in range(self.maxdim):
            latdframe, latdlblval, latdlbl, latdval, latd, latdtip = (
                add_spinbox(self.rowlatd, label=str(i), values=(0,), wrap=True,
                            command=self.spinned_lat, state=tk.DISABLED,
                            tooltip='None'))
            self.latdframe.append(latdframe)
            self.latdlblval.append(latdlblval)
            self.latdlbl.append(latdlbl)
            self.latdval.append(latdval)
            self.latd.append(latd)
            self.latdtip.append(latdtip)
            latdframe.pack(side=tk.LEFT)

        # 4. row
        # options
        self.rowcmap = Frame(self)
        self.rowcmap.pack(side=tk.TOP, fill=tk.X)
        self.cmapframe, self.cmaplbl, self.cmap, self.cmaptip = add_imagemenu(
            self.rowcmap, label='cmap', values=self.cmaps,
            images=self.imaps, command=self.selected_cmap,
            tooltip='Choose colormap')
        self.cmap['text']  = 'RdYlBu'
        self.cmap['image'] = self.imaps[self.cmaps.index('RdYlBu')]
        self.cmapframe.pack(side=tk.LEFT)
        self.rev_cmapframe, self.rev_cmaplbl, self.rev_cmap, self.rev_cmaptip = (
            add_checkbutton(self.rowcmap, label='reverse cmap', value=False,
                            command=self.checked,
                            tooltip='Reverse colormap'))
        self.rev_cmapframe.pack(side=tk.LEFT)
        self.meshframe, self.meshlbl, self.mesh, self.meshtip = (
            add_checkbutton(self.rowcmap, label='mesh', value=True,
                            command=self.checked,
                            tooltip=('Pseudocolor plot if checked,'
                                     ' contours if unchecked')))
        self.meshframe.pack(side=tk.LEFT)
        self.iglobalframe, self.igloballbl, self.iglobal, self.iglobaltip = (
            add_checkbutton(self.rowcmap, label='global', value=False,
                            command=self.checked,
                            tooltip='Assume global extent'))
        self.iglobalframe.pack(side=tk.LEFT)
        self.coastframe, self.coastlbl, self.coast, self.coasttip = (
            add_checkbutton(self.rowcmap, label='coast', value=True,
                            command=self.checked,
                            tooltip='Draw continental coast lines'))
        self.coastframe.pack(side=tk.LEFT)
        self.bordersframe, self.borderslbl, self.borders, self.borderstip = (
            add_checkbutton(self.rowcmap, label='borders', value=False,
                            command=self.checked,
                            tooltip='Draw country borders'))
        self.bordersframe.pack(side=tk.LEFT)
        self.riversframe, self.riverslbl, self.rivers, self.riverstip = (
            add_checkbutton(self.rowcmap, label='rivers', value=False,
                            command=self.checked,
                            tooltip='Draw rivers'))
        self.riversframe.pack(side=tk.LEFT)
        self.lakesframe, self.lakeslbl, self.lakes, self.lakestip = (
            add_checkbutton(self.rowcmap, label='lakes', value=False,
                            command=self.checked,
                            tooltip='Draw major lakes'))
        self.lakesframe.pack(side=tk.LEFT)
        self.gridframe, self.gridlbl, self.grid, self.gridtip = (
            add_checkbutton(self.rowcmap, label='grid', value=False,
                            command=self.checked,
                            tooltip='Draw major grid lines'))
        self.gridframe.pack(side=tk.LEFT)

        # 7. row
        # projections
        self.rowproj = Frame(self)
        self.rowproj.pack(side=tk.TOP, fill=tk.X)
        self.projframe, self.projlbl, self.proj, self.projtip = add_menu(
            self.rowproj, label='projection', values=self.projs,
            command=self.selected_proj, width=mwidth,
            tooltip='Choose projection')
        if ihavectk:
            self.proj.set('PlateCarree')
        else:
            self.proj['text'] = 'PlateCarree'
        self.projframe.pack(side=tk.LEFT)
        tstr = ('Central longitude of projection.\n'
                'Determined from longitude variable if None.')
        self.clonframe, self.clonlbl, self.clon, self.clontip = add_entry(
            self.rowproj, label='central lon', text='None', width=ewmed,
            command=self.entered_clon, tooltip=tstr, padx=padx)
        self.clonframe.pack(side=tk.LEFT)
        # Quit button
        self.bquit, self.bquittip = add_button(
            self.rowproj, text='Quit', command=self.master.top.destroy,
            nopack=True, tooltip='Quit ncvue')
        self.bquit.pack(side=tk.RIGHT)

        # set lat/lon
        if self.usex:
            if self.lonvar:
                self.lon.set(self.lonvar)
                self.inv_lon.set(0)
                self.shift_lon.set(0)
                set_dim_lon(self)
            if self.latvar:
                self.lat.set(self.latvar)
                self.inv_lat.set(0)
                set_dim_lat(self)
        else:
            if any(self.lonvar):
                idx = [ i for i, l in enumerate(self.lonvar) if l ]
                self.lon.set(self.lonvar[idx[0]])
                self.inv_lon.set(0)
                self.shift_lon.set(0)
                set_dim_lon(self)
            if any(self.latvar):
                idx = [ i for i, l in enumerate(self.latvar) if l ]
                self.lat.set(self.latvar[idx[0]])
                self.inv_lat.set(0)
                set_dim_lat(self)

        # set global
        x = self.lon.get()
        if (x != ''):
            gx, vx = vardim2var(x, self.groups)
            xx = selvar(self, vx)
            xx = get_slice_miss(self, self.lond, xx)
            if np.any(np.isfinite(xx)):
                xx = (xx + 360.) % 360.
            if (xx.max() - xx.min()) > 150.:
                self.iglobal.set(1)
            else:
                self.iglobal.set(0)

        # animation
        rep = self.repeat.get()
        if rep == 'repeat':
            irepeat = True
        else:
            irepeat = False
        self.anim_first   = True   # True: stops in self.update at first call
        self.anim_running = True   # True/False: animation running or not
        self.anim_inc     = 1      # 1/-1: forward or backward run
        maxtime = 1
        if self.usex:
            if self.tvar:
                zz = selvar(self, self.tvar)
                maxtime = max(zz.size, maxtime)
        else:
            for vz in self.tvar:
                if vz:
                    zz = selvar(self, vz)
                    maxtime = max(zz.size, maxtime)
        self.anim = animation.FuncAnimation(
            self.figure, self.update, init_func=self.redraw,
            interval=1, repeat=irepeat, save_count=maxtime)

    #
    # Bindings
    #

    def checked(self):
        """
        Command called if any checkbutton was checked or unchecked.

        Redraws plot.

        """
        self.redraw()

    def checked_all(self):
        """
        Command called if any checkbutton 'all' for vmin/vmax was checked or
        unchecked.

        Resets vmin/vmax, redraws plot.

        """
        vmin, vmax = self.get_vminmax()
        self.vmin.set(vmin)
        self.vmax.set(vmax)
        self.redraw()

    def entered_clon(self, event):
        """
        Command called if values central longitude was entered.

        Triggering `event` was bound to entry.

        Redraws plot.

        """
        self.redraw()

    def entered_v(self, event):
        """
        Command called if values for `vmin`/`vmax` were entered.

        Triggering `event` was bound to entry.

        Redraws plot.

        """
        self.redraw()

    def first_t(self):
        """
        Command called if first frame button was pressed.

        """
        it = 0
        self.set_tstep(it)
        self.update(it, isframe=True)

    def last_t(self):
        """
        Command called if last frame button was pressed.

        """
        it = self.nunlim - 1
        self.set_tstep(it)
        self.update(it, isframe=True)

    def newnetcdf(self):
        """
        Open a new netcdf file and connect it to top.

        """
        # get new netcdf file name
        ncfile = tk.filedialog.askopenfilename(
            parent=self, title='Choose netcdf file(s)', multiple=True)
        if len(ncfile) > 0:
            # close old netcdf file
            if self.usex:
                if self.top.fi:
                    self.top.fi.close()
            else:
                if len(self.top.fi) > 0:
                    for fi in self.top.fi:
                        fi.close()
            # reset empty defaults of top
            self.top.usex   = False  # open file with netCDF4 or xarray
            self.top.fi     = []  # file name or file handle
            self.top.groups = []  # filename with incr. index or group names
            self.top.dunlim = []  # name of unlimited dimension
            self.top.time   = []  # datetime variable
            self.top.tname  = []  # datetime variable name
            self.top.tvar   = []  # datetime variable name in netcdf
            self.top.dtime  = []  # decimal year
            self.top.latvar = []  # name of latitude variable
            self.top.lonvar = []  # name of longitude variable
            self.top.latdim = []  # name of latitude dimension
            self.top.londim = []  # name of longitude dimension
            self.top.maxdim = 0   # maximum num of dims of all variables
            self.top.cols   = []  # variable list
            # open new netcdf file
            ianalyse = True
            for ii, nn in enumerate(ncfile):
                self.top.fi.append(nc.Dataset(nn, 'r'))
                if len(ncfile) > 1:
                    nnc = np.ceil(np.log10(len(ncfile))).astype(int)
                    self.top.groups.append(f'file{ii:0{nnc}d}')
            # Check groups
            if len(ncfile) == 1:
                self.top.groups = list(self.top.fi[0].groups.keys())
            else:
                for ii, nn in enumerate(ncfile):
                    if len(list(self.top.fi[ii].groups.keys())) > 0:
                        print(f'Either multiple files or one file with'
                              f' groups allowed as input. Multiple files'
                              f' given but file {nn} has groups.')
                        for fi in self.top.fi:
                            fi.close()
                        ianalyse = False
            if ianalyse:
                analyse_netcdf(self.top)
            # reset panel
            self.reinit()
            self.redraw()

    def newxarray(self):
        """
        Open a new netcdf file and connect it to top.

        """
        # get new netcdf file name
        ncfile = tk.filedialog.askopenfilename(
            parent=self, title='Choose netcdf file(s)', multiple=True)
        if len(ncfile) > 0:
            # close old netcdf file
            if self.usex:
                if self.top.fi:
                    self.top.fi.close()
            else:
                if len(self.top.fi) > 0:
                    for fi in self.top.fi:
                        fi.close()
            # reset empty defaults of top
            self.top.usex   = True  # open file with netCDF4 or xarray
            self.top.fi     = []  # file name or file handle
            self.top.groups = []  # filename with incr. index or group names
            self.top.dunlim = []  # name of unlimited dimension
            self.top.time   = []  # datetime variable
            self.top.tname  = []  # datetime variable name
            self.top.tvar   = []  # datetime variable name in netcdf
            self.top.dtime  = []  # decimal year
            self.top.latvar = []  # name of latitude variable
            self.top.lonvar = []  # name of longitude variable
            self.top.latdim = []  # name of latitude dimension
            self.top.londim = []  # name of longitude dimension
            self.top.maxdim = 0   # maximum num of dims of all variables
            self.top.cols   = []  # variable list
            # open new netcdf file
            if len(ncfile) > 1:
                self.top.fi = xr.open_mfdataset(ncfile)
            else:
                self.top.fi = xr.open_dataset(ncfile[0])
            analyse_netcdf(self.top)
            # reset panel
            self.reinit()
            self.redraw()

    def nrun_t(self):
        """
        Command called if forward run button was pressed.

        """
        srun = self.nrun_timelbl.get()
        if srun == '>':
            if not self.anim_running:
                self.anim_inc = 1
                self.anim.event_source.start()
                self.anim_running = True
                self.nrun_timelbl.set('||')
        elif srun == '||':
            if self.anim_running:
                self.anim.event_source.stop()
                self.anim_running = False
                self.nrun_timelbl.set('>')

    def next_t(self):
        """
        Command called if next frame button was pressed.

        """
        try:
            it = int(self.vdval[self.iunlim].get())
        except ValueError:
            it = -1
        if (it < self.nunlim - 1) and (it >= 0):
            it += 1
            self.set_tstep(it)
            self.update(it, isframe=True)
        elif it == self.nunlim - 1:
            rep = self.repeat.get()
            if rep != 'once':
                if rep == 'repeat':
                    it = 0
                else:  # reflect
                    it -= 1
                self.set_tstep(it)
                self.update(it, isframe=True)

    def next_v(self):
        """
        Command called if next button for the plotting variable was pressed.

        Resets `vmin`/`vmax` and variable-dimensions. Redraws plot.

        """
        v = self.v.get()
        if ihavectk:
            cols = self.v.cget('values')
        else:
            cols = self.v['values']
        idx  = cols.index(v)
        idx += 1
        if idx < len(cols):
            self.v.set(cols[idx])
            self.set_unlim(cols[idx])
            if ihavectk:
                self.tstep.configure(to=self.nunlim - 1)
                from_ = self.tstep.cget('from_')
                to = self.tstep.cget('to')
                self.tstep.configure(number_of_steps=to - from_ + 1)
            else:
                self.tstep['to'] = self.nunlim - 1
            self.set_tstep(0)
            vmin, vmax = self.get_vminmax()
            self.vmin.set(vmin)
            self.vmax.set(vmax)
            set_dim_var(self)
            self.redraw()

    def prev_t(self):
        """
        Command called if previous frame button was pressed.

        """
        try:
            it = int(self.vdval[self.iunlim].get())
        except ValueError:
            it = -1
        if it > 0:
            it -= 1
            self.set_tstep(it)
            self.update(it, isframe=True)
        elif it == 0:
            rep = self.repeat.get()
            if rep != 'once':
                if rep == 'repeat':
                    it = self.nunlim - 1
                else:  # reflect
                    it += 1
                self.set_tstep(it)
                self.update(it, isframe=True)

    def prev_v(self):
        """
        Command called if previous button for the plotting variable was
        pressed.

        Resets `vmin`/`vmax` and v-dimensions. Redraws plot.

        """
        v = self.v.get()
        if ihavectk:
            cols = self.v.cget('values')
        else:
            cols = self.v['values']
        idx  = cols.index(v)
        idx -= 1
        if idx > 0:
            self.v.set(cols[idx])
            self.set_unlim(cols[idx])
            if ihavectk:
                self.tstep.configure(to=self.nunlim - 1)
                from_ = self.tstep.cget('from_')
                to = self.tstep.cget('to')
                self.tstep.configure(number_of_steps=to - from_ + 1)
            else:
                self.tstep['to'] = self.nunlim - 1
            self.set_tstep(0)
            vmin, vmax = self.get_vminmax()
            self.vmin.set(vmin)
            self.vmax.set(vmax)
            set_dim_var(self)
            self.redraw()

    def prun_t(self):
        """
        Command called if run backwards button was pressed.

        """
        srun = self.prun_timelbl.get()
        if srun == '<':
            if not self.anim_running:
                self.anim_inc = -1
                self.anim.event_source.start()
                self.anim_running = True
                self.prun_timelbl.set('||')
        elif srun == '||':
            if self.anim_running:
                self.anim.event_source.stop()
                self.anim_running = False
                self.prun_timelbl.set('<')

    def repeat_t(self, event):
        """
        Command called if repeat option was chosen with combobox.

        Triggering `event` was bound to the combobox.

        """
        rep = self.repeat.get()
        if rep == 'once':
            irepeat = False
        else:
            # need not to stop also for reflect
            irepeat = True
        self.anim.repeat = irepeat

    def selected_cmap(self, value):
        """
        Command called if cmap was chosen from menu.

        `value` is the chosen colormap.

        Sets text and image on the menubutton.

        """
        self.cmap['text']  = value
        self.cmap['image'] = self.imaps[self.cmaps.index(value)]
        self.redraw()

    def selected_lat(self, event):
        """
        Command called if latitude-variable was selected with combobox.

        Triggering `event` was bound to the combobox.

        Resets `lat` options and dimensions. Redraws plot.

        """
        self.inv_lat.set(0)
        set_dim_lat(self)
        self.redraw()

    def selected_lon(self, event):
        """
        Command called if x-variable was selected with combobox.

        Triggering `event` was bound to the combobox.

        Resets `x` options and dimensions. Redraws plot.

        """
        self.inv_lon.set(0)
        self.shift_lon.set(0)
        set_dim_lon(self)
        self.redraw()

    def selected_proj(self, value):
        """
        Command called if projection was chosen from menu.

        `value` is the chosen projection.

        Sets text on the menubutton.

        """
        if ihavectk:
            self.proj.set(value)
        else:
            self.proj['text'] = value
        self.redraw()

    def selected_v(self, event):
        """
        Command called if plotting variable was selected with combobox.

        Triggering `event` was bound to the combobox.

        Resets `vmin`/`vmax` and variable-dimensions. Redraws plot.

        """
        v = self.v.get()
        self.set_unlim(v)
        if ihavectk:
            self.tstep.configure(to=self.nunlim - 1)
            from_ = self.tstep.cget('from_')
            to = self.tstep.cget('to')
            self.tstep.configure(number_of_steps=to - from_ + 1)
        else:
            self.tstep['to'] = self.nunlim - 1
        self.set_tstep(0)
        vmin, vmax = self.get_vminmax()
        self.vmin.set(vmin)
        self.vmax.set(vmax)
        set_dim_var(self)
        self.redraw()

    def spinned_lon(self, event=None):
        """
        Command called if spinbox of x-dimensions was changed.

        Triggering `event` was bound to the spinbox.

        Redraws plot.

        """
        self.redraw()

    def spinned_lat(self, event=None):
        """
        Command called if spinbox of latitude-dimensions was changed.

        Triggering `event` was bound to the spinbox.

        Redraws plot.

        """
        self.redraw()

    def spinned_v(self, event=None):
        """
        Command called if spinbox of variable-dimensions was changed.

        Triggering `event` was bound to the spinbox.

        Redraws plot.

        """
        try:
            it = int(self.vdval[self.iunlim].get())
            self.set_tstep(it)
        except ValueError:  # mean, std, etc.
            pass
        self.redraw()

    def tstep_t(self, step):
        """
        Command called if tstep was changed.

        `step` is the chosen value on the scale slider.

        """
        it = int(float(step))
        self.set_tstep(it)
        self.update(it, isframe=True)

    #
    # Methods
    #

    def get_vminmax(self):
        from numpy.random import default_rng
        v = self.v.get()
        if (v != ''):
            gz, vz = vardim2var(v, self.groups)
            if self.usex:
                tname = self.tname
            else:
                tname = self.tname[gz]
            if vz == tname:
                return (0, 1)
            vv = selvar(self, vz)
            imiss = get_miss(self, vv)
            iall  = self.vall.get()
            if iall or (np.sum(vv.shape[:-2]) < 50):
                vv   = set_miss(imiss, vv)
                vmin = np.nanmin(vv)
                vmax = np.nanmax(vv)
            else:
                rng = default_rng()
                vmin = np.inf
                vmax = -np.inf
                for nn in range(50):
                    ss = []
                    for i in range(vv.ndim):
                        if i < vv.ndim - 2:
                            idim = rng.integers(0, vv.shape[i])
                            s = slice(idim, idim + 1)
                        else:
                            s = slice(0, vv.shape[i])
                        ss.append(s)
                    ivv   = vv[tuple(ss)]
                    ivv   = set_miss(imiss, ivv)
                    ivmin = np.nanmin(ivv)
                    ivmax = np.nanmax(ivv)
                    vmin  = min(vmin, ivmin)
                    vmax  = max(vmax, ivmax)
            return (vmin, vmax)
        else:
            return (0, 1)

    def reinit(self):
        """
        Reinitialise the panel from top.

        """
        # reinit from top
        self.usex   = self.top.usex
        self.fi     = self.top.fi
        self.groups = self.top.groups
        self.miss   = self.top.miss
        self.dunlim = self.top.dunlim
        self.time   = self.top.time
        self.tname  = self.top.tname
        self.tvar   = self.top.tvar
        self.dtime  = self.top.dtime
        self.latvar = self.top.latvar
        self.lonvar = self.top.lonvar
        self.latdim = self.top.latdim
        self.londim = self.top.londim
        self.maxdim = self.top.maxdim
        self.cols   = self.top.cols
        self.iunlim = -1
        self.nunlim = 0
        # reset dimensions
        for ll in self.vdlbl:
            ll.destroy()
        for ll in self.vd:
            ll.destroy()
        for ll in self.vdframe:
            ll.destroy()
        self.vdframe  = []
        self.vdlblval = []
        self.vdlbl    = []
        self.vdval    = []
        self.vd       = []
        self.vdtip    = []
        for i in range(self.maxdim):
            vdframe, vdlblval, vdlbl, vdval, vd, vdtip = add_spinbox(
                self.rowvd, label=str(i), values=(0,), wrap=True,
                command=self.spinned_v, state=tk.DISABLED, tooltip='None')
            self.vdframe.append(vdframe)
            self.vdlblval.append(vdlblval)
            self.vdlbl.append(vdlbl)
            self.vdval.append(vdval)
            self.vd.append(vd)
            self.vdtip.append(vdtip)
            vdframe.pack(side=tk.LEFT)
        for ll in self.latdlbl:
            ll.destroy()
        for ll in self.latd:
            ll.destroy()
        for ll in self.latdframe:
            ll.destroy()
        self.latdframe  = []
        self.latdlblval = []
        self.latdlbl    = []
        self.latdval    = []
        self.latd       = []
        self.latdtip    = []
        for i in range(self.maxdim):
            latdframe, latdlblval, latdlbl, latdval, latd, latdtip = (
                add_spinbox(self.rowlatd, label=str(i), values=(0,), wrap=True,
                            command=self.spinned_lat, state=tk.DISABLED,
                            tooltip='None'))
            self.latdframe.append(latdframe)
            self.latdlblval.append(latdlblval)
            self.latdlbl.append(latdlbl)
            self.latdval.append(latdval)
            self.latd.append(latd)
            self.latdtip.append(latdtip)
            latdframe.pack(side=tk.LEFT)
        for ll in self.londlbl:
            ll.destroy()
        for ll in self.lond:
            ll.destroy()
        for ll in self.londframe:
            ll.destroy()
        self.londframe  = []
        self.londlblval = []
        self.londlbl    = []
        self.londval    = []
        self.lond       = []
        self.londtip    = []
        for i in range(self.maxdim):
            londframe, londlblval, londlbl, londval, lond, londtip = (
                add_spinbox(self.rowlond, label=str(i), values=(0,), wrap=True,
                            command=self.spinned_lon, state=tk.DISABLED,
                            tooltip='None'))
            self.londframe.append(londframe)
            self.londlblval.append(londlblval)
            self.londlbl.append(londlbl)
            self.londval.append(londval)
            self.lond.append(lond)
            self.londtip.append(londtip)
            londframe.pack(side=tk.LEFT)
        # set time step
        if ihavectk:
            self.tstep.configure(to=1)
            from_ = self.tstep.cget('from_')
            to = self.tstep.cget('to')
            self.tstep.configure(number_of_steps=to - from_ + 1)
        else:
            self.tstep['to'] = 1
        self.tstepval.set(0)
        self.repeat.set('repeat')
        # set variables
        columns = [''] + self.cols
        if ihavectk:
            self.v.configure(values=columns)
        else:
            self.v['values'] = columns
        self.v.set(columns[0])
        self.vmin.set('None')
        self.vmax.set('None')
        # set lat/lon
        if ihavectk:
            self.lon.configure(values=columns)
        else:
            self.lon['values'] = columns
        self.lon.set(columns[0])
        if ihavectk:
            self.lat.configure(values=columns)
        else:
            self.lat['values'] = columns
        self.lat.set(columns[0])
        if self.usex:
            if self.latvar:
                self.lat.set(self.latvar)
                self.inv_lat.set(0)
                set_dim_lat(self)
            if self.lonvar:
                self.lon.set(self.lonvar)
                self.inv_lon.set(0)
                self.shift_lon.set(0)
                set_dim_lon(self)
        else:
            if any(self.latvar):
                idx = [ i for i, l in enumerate(self.latvar) if l ]
                self.lat.set(self.latvar[idx[0]])
                self.inv_lat.set(0)
                set_dim_lat(self)
            if any(self.lonvar):
                idx = [ i for i, l in enumerate(self.lonvar) if l ]
                self.lon.set(self.lonvar[idx[0]])
                self.inv_lon.set(0)
                self.shift_lon.set(0)
                set_dim_lon(self)
        x = self.lon.get()
        if (x != ''):
            gx, vx = vardim2var(x, self.groups)
            xx = selvar(self, vx)
            xx = get_slice_miss(self, self.lond, xx)
            xx = xx + 360.
            if (xx.max() - xx.min()) > 150.:
                self.iglobal.set(1)
            else:
                self.iglobal.set(0)

    def set_tstep(self, it):
        """
        Make all steps when changing time step.

        `it` (int) is the time step.

        Sets the time dimension spinbox, sets the time step scale,
        write the time on top.

        """
        v = self.v.get()
        gz, vz = vardim2var(v, self.groups)
        if self.usex:
            dunlim = self.dunlim
            time = self.time
        else:
            dunlim = self.dunlim[gz]
            time = self.time[gz]
        try:
            zz = selvar(self, vz)
            if self.usex:
                dims = zz.dims
            else:
                dims = zz.dimensions
            has_unlim = dunlim in dims
        except IndexError:
            has_unlim = False  # datetime
        if dunlim and has_unlim:
            self.vdval[self.iunlim].set(it)
            self.tstepval.set(it)
            if self.usex:
                self.timelbl.set(str(time.values[it]))
            else:
                try:
                    self.timelbl.set(np.around(time[it], 4))
                except TypeError:
                    self.timelbl.set(time[it])

    def set_unlim(self, v):
        """
        Set index and length of unlimited dimension of variable `v`.

        `v` (str) is the variable name as in the selection comboboxes, i.e.
        `gvar, var = vardim2var(self.fi[v)]`.

        Sets `self.nunlim` to the length of the unlimited dimension and
        `self.iunlim` to the index in variable.dimensions if
        `self.dunlim ~= ''` and `self.dunlim` in var.dimensions.

        Takes `self.iunlim=0` and `self.nunlim=variable.shape[0]` if
        self.dunlim == ''` or `self.dunlim` not in var.dimensions.

        """
        gz, vz = vardim2var(v, self.groups)
        if self.usex:
            tname = self.tname
            time = self.time
            dunlim = self.dunlim
        else:
            tname = self.tname[gz]
            time = self.time[gz]
            dunlim = self.dunlim[gz]
        if vz == tname:
            self.iunlim = 0
            self.nunlim = time.size
        else:
            zz = selvar(self, vz)
            if self.usex:
                dims = zz.dims
            else:
                dims = zz.dimensions
            if dunlim:
                if dunlim in dims:
                    self.iunlim = dims.index(dunlim)
                else:
                    self.iunlim = 0
            else:
                self.iunlim = 0
            if zz.ndim > 0:
                self.nunlim = zz.shape[self.iunlim]
            else:
                self.nunlim = 0

    #
    # Plotting
    #

    # def redraw(self):
    #     pass

    def redraw(self):
        """
        Redraws the plot.

        Reads `lon`, `lat`, `variable` names, the current settings of
        their dimension spinboxes, as well as all other plotting options.
        Then redraws the plot.

        """
        # stop animation
        self.anim.event_source.stop()
        self.anim_running = False
        # get all states
        # rowv
        v = self.v.get()
        trans_v = self.trans_v.get()
        vmin = self.vmin.get()
        if vmin == 'None':
            vmin = None
        else:
            try:
                vmin = float(vmin)
            except ValueError:
                vmin = None
        vmax = self.vmax.get()
        if vmax == 'None':
            vmax = None
        else:
            try:
                vmax = float(vmax)
            except ValueError:
                vmax = None
        # rowll
        x = self.lon.get()
        y = self.lat.get()
        inv_lon   = self.inv_lon.get()
        inv_lat   = self.inv_lat.get()
        shift_lon = self.shift_lon.get()
        # rowcmap
        cmap     = self.cmap['text']
        rev_cmap = self.rev_cmap.get()
        mesh     = self.mesh.get()
        self.iiglobal = self.iglobal.get()
        coast    = self.coast.get()
        borders  = self.borders.get()
        rivers   = self.rivers.get()
        lakes    = self.lakes.get()
        grid     = self.grid.get()
        if ihavectk:
            proj = self.proj.get()
        else:
            proj = self.proj['text']
        self.iproj = self.iprojs[self.projs.index(proj)]
        clon     = self.clon.get()
        # set x, y, axes labels
        vx = 'None'
        vy = 'None'
        vz = 'None'
        if (v != ''):
            # variable
            gz, vz = vardim2var(v, self.groups)
            if self.usex:
                tnamez = self.tname
                dtimez = self.dtime
                timez = self.time
            else:
                tnamez = self.tname[gz]
                dtimez = self.dtime[gz]
                timez = self.time[gz]
            if vz == tnamez:
                # should throw an error later
                if mesh:
                    vv = dtimez
                    vlab = 'Year'
                else:
                    vv = timez
                    vlab = 'Date'
            else:
                vv = selvar(self, vz)
                vlab = set_axis_label(vv)
            vv = get_slice_miss(self, self.vd, vv)
            if trans_v:
                vv = vv.T
            if shift_lon:
                vv = np.roll(vv, vv.shape[1] // 2, axis=1)
        else:
            vlab = ''
        if (y != ''):
            # y axis
            gy, vy = vardim2var(y, self.groups)
            if self.usex:
                tnamey = self.tname
                dtimey = self.dtime
                timey = self.time
            else:
                tnamey = self.tname[gy]
                dtimey = self.dtime[gy]
                timey = self.time[gy]
            if vy == tnamey:
                if mesh:
                    yy = dtimey
                    ylab = 'Year'
                else:
                    yy = timey
                    ylab = 'Date'
            else:
                yy = selvar(self, vy)
                ylab = set_axis_label(yy)
            yy = get_slice_miss(self, self.latd, yy)
        else:
            ylab = ''
        if (x != ''):
            # x axis
            gx, vx = vardim2var(x, self.groups)
            if self.usex:
                tnamex = self.tname
                dtimex = self.dtime
                timex = self.time
            else:
                tnamex = self.tname[gx]
                dtimex = self.dtime[gx]
                timex = self.time[gx]
            if vx == tnamex:
                if mesh:
                    xx = dtimex
                    xlab = 'Year'
                else:
                    xx = timex
                    xlab = 'Date'
            else:
                xx = selvar(self, vx)
                xlab = set_axis_label(xx)
            xx = get_slice_miss(self, self.lond, xx)
            # set central longitude of projection
            # make in 0-360, otherwise always 0 if -180 to 180
            if np.any(np.isfinite(xx)):
                xx360 = (xx + 360.) % 360.
            else:
                xx360 = xx
            if xx.size > 1:
                if xx.ndim > 1:
                    x0 = xx[:, 0].mean()
                    if self.iiglobal:
                        # -2 instead of -1 to avoid possible cyclic longitude
                        x1 = xx[:, -2].mean()
                    else:
                        x1 = xx[:, -1].mean()
                else:
                    x0 = xx[0]
                    if self.iiglobal:
                        x1 = xx[-2]
                    else:
                        x1 = xx[-1]
                self.ixxmean = 0.5 * (x1 + x0)
                if self.iiglobal:
                    # round it to next 180 degrees to get 0 or 180
                    self.ixxmean = np.around(self.ixxmean / 180., 0) * 180.
            else:
                self.ixxmean = xx360[0]
            # seems to work better if central lon in projection
            # is set from -180 to 180 even if lon is given in 0-360
            if self.ixxmean > 180.:
                self.ixxmean -= 360.
        else:
            xlab = ''
            self.ixxmean = 0.
        if clon != 'None':
            self.iclon = float(clon)
        else:
            self.iclon = self.ixxmean
        # plot options
        if rev_cmap:
            cmap = cmap + '_r'
        # Clear figure instead of axes because colorbar is on figure
        self.figure.clear()
        # Have to add axes again.
        self.axes = self.figure.add_subplot(
            111, projection=self.iproj(central_longitude=self.iclon))
        # plot only if variable given
        if (v != ''):
            if vv.ndim < 2:
                print(f'Map: var ({vz}) is not 2-dimensional:', vv.shape)
                return
            # set x and y to index if not selected
            if (x == ''):
                nx = vv.shape[1]
                xx = -180. + np.arange(nx) / float(nx) * 360.
                xx += 0.5 * (xx[1] - xx[0])
                xlab = ''
            if (y == ''):
                ny = vv.shape[0]
                yy = -90. + np.arange(ny) / float(ny) * 180.
                yy += 0.5 * (yy[1] - yy[0])
                ylab = ''
            # plot
            # cc = self.axes.imshow(vv[:, ::-1].T, aspect='auto', cmap=cmap,
            #                       interpolation='none')
            # cc = self.axes.matshow(vv[:, ::-1].T, aspect='auto', cmap=cmap,
            #                        interpolation='none')
            extend = 'neither'
            if vmin is not None:
                vv = np.maximum(vv, vmin)
                if vmax is None:
                    extend = 'min'
                else:
                    extend = 'both'
            if vmax is not None:
                vv = np.minimum(vv, vmax)
                if vmin is None:
                    extend = 'max'
                else:
                    extend = 'both'
            if (xx.ndim == 1) and (yy.ndim == 1):
                self.ixx, self.iyy = np.meshgrid(xx, yy)
            elif (xx.ndim == 1) and (yy.ndim == 2):
                self.ixx, tmp = np.meshgrid(xx, yy[:, 0])
                self.iyy = yy
            elif (xx.ndim == 2) and (yy.ndim == 1):
                self.ixx, self.iyy = np.meshgrid(xx, yy)
                tmp, self.iyy = np.meshgrid(xx[0, :], yy)
                self.ixx = xx
            elif (xx.ndim == 2) and (yy.ndim == 2):
                self.ixx = xx
                self.iyy = yy
            else:
                print(f'Map: lon ({vx}), lat ({vy}) dimensions not 1D or 2D:',
                      xx.shape, yy.shape)
                return
            if inv_lon:
                self.ixx = np.fliplr(self.ixx)
            if inv_lat:
                self.iyy = np.flipud(self.iyy)
            self.ivv = vv
            if self.iiglobal:
                # cartopy.contourf needs cyclic longitude for wrap around
                self.ivvc, self.ixxc, self.iyyc = add_cyclic(
                    self.ivv, x=self.ixx, y=self.iyy)
                # # special treatment if fringe points < 1e-4 apart
                # # This works but it did still not display correctly the
                # # test file of GDPS from cuizinart
                # # -> do not do it for the moment
                # if np.ma.allclose(np.ma.sin(np.deg2rad(self.ixxc[:, 0])),
                #                   np.ma.sin(np.deg2rad(self.ixxc[:, -1])),
                #                   atol=1.0e-5):
                #     if np.ma.allclose(self.ixxc[:, 0], self.ixxc[:, -1],
                #                       atol=1.0e-4):
                #         self.ixxc = self.ixxc.astype(float)
                #         self.ixxc[:, -1] = self.ixxc[:, 0] + 360.
            else:
                self.ivvc = self.ivv
                self.ixxc = self.ixx
                self.iyyc = self.iyy
            self.itrans  = ccrs.PlateCarree()
            self.ivmin   = vmin
            self.ivmax   = vmax
            self.icmap   = cmap
            self.ncmap   = mpl.colormaps[self.icmap].N
            self.ncmap   = self.ncmap if self.ncmap < 256 else 15
            self.iextend = extend
            if mesh:
                try:
                    # vv is matrix notation: (row, col)
                    # self.cc = self.axes.pcolormesh(
                    #     xx, yy, vv, vmin=vmin, vmax=vmax, cmap=cmap,
                    #     shading='nearest')
                    self.cc = self.axes.pcolormesh(
                        self.ixx, self.iyy, self.ivv,
                        vmin=self.ivmin, vmax=self.ivmax,
                        cmap=self.icmap, shading='nearest',
                        transform=self.itrans)
                    # self.cc = self.axes.imshow(
                    #     vv, vmin=vmin, vmax=vmax, cmap=cmap,
                    #     origin='upper', extent=self.img_extent,
                    #     transform=self.itrans)
                    self.cb = self.figure.colorbar(self.cc, fraction=0.05,
                                                   shrink=0.75, pad=0.07,
                                                   extend=self.iextend)
                except Exception:
                    print(f'Map pcolormesh: lon ({vx}), lat ({vy}),'
                          f' var ({vz}) shapes do not match:',
                          self.ixx.shape, self.iyy.shape, self.ivv.shape)
                    return
            else:
                try:
                    # if 1-D then len(x)==m (columns) and
                    #     len(y)==n (rows): v(n,m)
                    self.cc = self.axes.contourf(
                        self.ixxc, self.iyyc, self.ivvc, self.ncmap,
                        vmin=self.ivmin, vmax=self.ivmax,
                        cmap=self.icmap, extend=self.iextend,
                        transform=self.itrans)
                    self.cb = self.figure.colorbar(self.cc, fraction=0.05,
                                                   shrink=0.75, pad=0.07)
                    # self.cc, = self.axes.plot(yy, vv[0,:])
                except Exception:
                    print('Map contourf: lon ({vx}), lat ({vy}),'
                          f' var ({vz}) shapes do not match for:',
                          self.ixxc.shape, self.iyyc.shape, self.ivvc.shape)
                    return
            self.cb.set_label(vlab)
            self.axes.format_coord = lambda x, y: format_coord_map(
                x, y, self.axes, self.ixx, self.iyy, self.ivv)
        # help(self.figure)
        if self.iiglobal:
            self.axes.set_global()
        if coast:
            # self.axes.coastlines()
            self.axes.add_feature(cfeature.COASTLINE)
            self.axes.gridlines(draw_labels=True, linewidth=0,
                                x_inline=False, y_inline=False)
        if borders:
            self.axes.add_feature(cfeature.BORDERS, edgecolor='grey')
        if rivers:
            self.axes.add_feature(cfeature.RIVERS)
        if lakes:
            self.axes.add_feature(cfeature.LAKES, alpha=0.5)
        self.axes.xaxis.set_label_text(xlab)
        self.axes.yaxis.set_label_text(ylab)
        if grid:
            self.axes.gridlines(draw_labels=False,
                                x_inline=False, y_inline=False)
        # redraw
        self.canvas.draw()
        self.toolbar.update()

    # def update(self, frame, isframe=False):
    #     pass

    def update(self, frame, isframe=False):
        """
        Updates data of the current plot.

        """
        if self.anim_first:
            self.anim.event_source.stop()
            self.anim_running = False
            self.anim_first   = False
            return
        # variable
        v = self.v.get()
        if (v != ''):
            trans_v = self.trans_v.get()
            mesh = self.mesh.get()
            rep = self.repeat.get()
            # inv_lon = self.inv_lon.get()
            # inv_lat = self.inv_lat.get()
            shift_lon = self.shift_lon.get()
            gz, vz = vardim2var(v, self.groups)
            if self.usex:
                if vz == self.tname:
                    vz = self.tvar
            else:
                if vz == self.tname[gz]:
                    vz = self.tvar[gz]
            vv = selvar(self, vz)
            # slice
            try:
                it = int(self.vdval[self.iunlim].get())
                if not isframe:
                    if (self.anim_inc == 1) and (it == self.nunlim - 1):
                        if rep == 'repeat':
                            it = 0
                        elif rep == 'reflect':
                            self.anim_inc = -1
                            it += self.anim_inc
                        else:  # once
                            self.anim.event_source.stop()
                            self.anim_running = False
                    elif (self.anim_inc == -1) and (it == 0):
                        if rep == 'repeat':
                            it = self.nunlim - 1
                        elif rep == 'reflect':
                            self.anim_inc = 1
                            it += self.anim_inc
                        else:  # once
                            self.anim.event_source.stop()
                            self.anim_running = False
                    else:
                        it += self.anim_inc
            except ValueError:
                it = 0
            self.set_tstep(it)
            vv = get_slice_miss(self, self.vd, vv)
            if vv.ndim < 2:
                self.anim.event_source.stop()
                self.anim_running = False
                return
            if trans_v:
                vv = vv.T
            if shift_lon:
                vv = np.roll(vv, vv.shape[1] // 2, axis=1)
            self.ivv = vv
            # set data
            if mesh:
                # update works well on 'normal' pcolormesh but not on Cartopy's
                # self.cc.set_array(vv)
                # Both, imshow and pcolormesh need to remove the old
                # image.AxesImage or collections.QuadMesh first and then redraw
                # because the set_data (imshow) and set_array (pcolormesh) do
                # not respect transformations.
                self.cc.remove()
                self.cc = self.axes.pcolormesh(
                    self.ixx, self.iyy, self.ivv,
                    vmin=self.ivmin, vmax=self.ivmax,
                    cmap=self.icmap, shading='nearest',
                    transform=self.itrans)
                # self.cc.remove()
                # self.cc = self.axes.imshow(
                #     vv, vmin=self.ivmin, vmax=self.ivmax, cmap=self.icmap,
                #     origin='upper', extent=self.img_extent,
                #     transform=self.itrans)
            else:
                # # https://discourse.matplotlib.org/t/update-an-existing-contour-plot-with-new-data/13935
                # for coll in self.cc.collections:
                #     self.axes.collections.remove(coll)
                self.cc.remove()
                if self.iiglobal:
                    # self.ivvc = add_cyclic(self.ivv)
                    self.ivvc, self.ixxc = add_cyclic(
                        self.ivv, x=self.ixx)
                else:
                    self.ivvc = self.ivv
                self.cc = self.axes.contourf(
                    self.ixxc, self.iyyc, self.ivvc, self.ncmap,
                    vmin=self.ivmin, vmax=self.ivmax,
                    cmap=self.icmap, extend=self.iextend,
                    transform=self.itrans)
            self.canvas.draw_idle()
            return self.cc,



#!/usr/bin/env python
"""
Utility functions for ncvue.

The utility functions do not depend on the ncvue class.
Functions depending on the class are in ncvmethods.

This module was written by Matthias Cuntz while at Institut National de
Recherche pour l'Agriculture, l'Alimentation et l'Environnement (INRAE), Nancy,
France.

:copyright: Copyright 2020- Matthias Cuntz - mc (at) macu (dot) de
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following functions are provided:

.. autosummary::
   DIMMETHODS
   add_cyclic
   clone_ncvmain
   format_coord_contour
   format_coord_map
   format_coord_scatter
   get_slice
   get_standard_name
   get_units
   list_intersection
   selvar
   set_axis_label
   set_miss
   spinbox_values
   vardim2var
   xzip_dim_name_length
   zip_dim_name_length

History
   * Written Nov-Dec 2020 by Matthias Cuntz (mc (at) macu (dot) de)
   * General get_slice function from individual methods for x, y, y2, z,
     Dec 2020, Matthias Cuntz
   * Added arithmetics to apply on axis/dimensions such as mean, std, etc.,
     Dec 2020, Matthias Cuntz
   * Added clone_ncvmain, removing its own module, Dec 2020, Matthias Cuntz
   * Added SEPCHAR and DIMMETHODS, Jan 2021, Matthias Cuntz
   * Pass only ncvMain widget to clone_ncvmain, Jan 2021, Matthias Cuntz
   * Pass only root widget to clone_ncvmain, Jan 2021, Matthias Cuntz
   * Set correct missing value for date variable in numpy's datetime64[ms]
     format May 2021, Matthias Cuntz
   * Added format_coord functions for scatter, contour, and map,
     May 2021, Matthias Cuntz
   * Replaced add_cyclic_point with add_cyclic as submitted to cartopy,
     Jun 2021, Matthias Cuntz
   * Removed SEPCHAR, Jun 2021, Matthias Cuntz
   * Final add_cyclic, has_cyclic from cartopy v0.20.1,
     Nov 2021, Matthias Cuntz
   * Address fi.variables[name] directly by fi[name], Jan 2024, Matthias Cuntz
   * Add selvar to allow multiple netcdf files, Jan 2024, Matthias Cuntz
   * Remove [ms] from check for datetime in format_coord on axes,
     Oct 2024, Matthias Cuntz
   * Use CustomTkinter in clone_ncvmain if installed, Nov 2024, Matthias Cuntz
   * Increased digits in format_coord_scatter, Jan 2025, Matthias Cuntz
   * Increased digits in format_coord_contour and format_coord_map,
     Jan 2025, Matthias Cuntz
   * Use f-string in zip_dim_name_length, Feb 2025, Matthias Cuntz
   * Added xzip_dim_name_length, get_standard_name, get_units,
     Feb 2025, Matthias Cuntz
   * Allow xarray in selvar, Feb 2025, Matthias Cuntz

"""
import tkinter as tk
try:
    from customtkinter import CTkToplevel as Toplevel
except ModuleNotFoundError:
    from tkinter import Toplevel
import numpy as np
import matplotlib.dates as mpld
import cartopy.crs as ccrs
try:
    import xarray as xr
    ihavex = True
except ModuleNotFoundError:
    ihavex = False
import ncvue


__all__ = ['DIMMETHODS',
           'add_cyclic', 'has_cyclic', 'clone_ncvmain',
           'format_coord_contour', 'format_coord_map', 'format_coord_scatter',
           'get_slice', 'get_standard_name', 'get_units',
           'list_intersection', 'selvar', 'set_axis_label', 'set_miss',
           'spinbox_values', 'vardim2var',
           'xzip_dim_name_length', 'zip_dim_name_length']


DIMMETHODS = ('mean', 'std', 'min', 'max', 'ptp', 'sum', 'median', 'var')
"""
Arithmetic methods implemented on dimensions.

mean - average
std - standard deviation
min - minimum
max - maximum
ptp - point-to-point amplitude = max - min
sum - sum
median - 50-percentile
var - variance

"""


def _add_cyclic_data(data, axis=-1):
    """
    Add a cyclic point to a data array.

    Parameters
    ----------
    data : ndarray
        An n-dimensional array of data to add a cyclic point to.
    axis : int, optional
        Specifies the axis of the data array to add the cyclic point to.
        Defaults to the right-most axis.

    Returns
    -------
    The data array with a cyclic point added.

    """
    slicer = [slice(None)] * data.ndim
    try:
        slicer[axis] = slice(0, 1)
    except IndexError:
        raise ValueError(
            'The specified axis does not correspond to an array dimension.')
    npc = np.ma if np.ma.is_masked(data) else np
    return npc.concatenate((data, data[tuple(slicer)]), axis=axis)


def _add_cyclic_x(x, axis=-1, cyclic=360):
    """
    Add a cyclic point to a x/longitude coordinate array.

    Parameters
    ----------
    x : ndarray
        An array which specifies the x-coordinate values for
        the dimension the cyclic point is to be added to.
    axis : int, optional
        Specifies the axis of the x-coordinate array to add the cyclic point
        to. Defaults to the right-most axis.
    cyclic : float, optional
        Width of periodic domain (default: 360)

    Returns
    -------
    The coordinate array ``x`` with a cyclic point added.

    """
    npc = np.ma if np.ma.is_masked(x) else np
    # get cyclic x-coordinates
    # cx is the code from basemap (addcyclic)
    # https://github.com/matplotlib/basemap/blob/master/lib/mpl_toolkits/basemap/__init__.py
    cx = (np.take(x, [0], axis=axis) +
          cyclic * np.sign(np.diff(np.take(x, [0, -1], axis=axis),
                                   axis=axis)))
    # basemap ensures that the values do not exceed cyclic
    # (next code line). We do not do this to deal with rotated grids that
    # might have values not exactly 0.
    #     cx = npc.where(cx <= cyclic, cx, np.mod(cx, cyclic))
    return npc.concatenate((x, cx), axis=axis)


def has_cyclic(x, axis=-1, cyclic=360, precision=1e-4):
    """
    Check if x/longitude coordinates already have a cyclic point.

    Checks all differences between the first and last
    x-coordinates along ``axis`` to be less than ``precision``.

    Parameters
    ----------
    x : ndarray
        An array with the x-coordinate values to be checked for cyclic points.
    axis : int, optional
        Specifies the axis of the ``x`` array to be checked.
        Defaults to the right-most axis.
    cyclic : float, optional
        Width of periodic domain (default: 360).
    precision : float, optional
        Maximal difference between first and last x-coordinate to detect
        cyclic point (default: 1e-4).

    Returns
    -------
    True if a cyclic point was detected along the given axis,
    False otherwise.

    Examples
    --------
    Check for cyclic x-coordinate in one dimension.
    >>> import numpy as np
    >>> lons = np.arange(0, 360, 60)
    >>> clons = np.arange(0, 361, 60)
    >>> print(has_cyclic(lons))
    False
    >>> print(has_cyclic(clons))
    True

    Check for cyclic x-coordinate in two dimensions.
    >>> lats = np.arange(-90, 90, 30)
    >>> lon2d, lat2d = np.meshgrid(lons, lats)
    >>> clon2d, clat2d = np.meshgrid(clons, lats)
    >>> print(has_cyclic(lon2d))
    False
    >>> print(has_cyclic(clon2d))
    True

    """
    npc = np.ma if np.ma.is_masked(x) else np
    # transform to 0-cyclic, assuming e.g. -180 to 180 if any < 0
    x1 = np.mod(npc.where(x < 0, x + cyclic, x), cyclic)
    dd = np.diff(np.take(x1, [0, -1], axis=axis), axis=axis)
    if npc.all(np.abs(dd) < precision):
        return True
    else:
        return False


def add_cyclic(data, x=None, y=None, axis=-1,
               cyclic=360, precision=1e-4):
    """
    Add a cyclic point to an array and optionally corresponding
    x/longitude and y/latitude coordinates.

    Checks all differences between the first and last
    x-coordinates along ``axis`` to be less than ``precision``.

    Parameters
    ----------
    data : ndarray
        An n-dimensional array of data to add a cyclic point to.
    x : ndarray, optional
        An n-dimensional array which specifies the x-coordinate values
        for the dimension the cyclic point is to be added to, i.e. normally the
        longitudes. Defaults to None.

        If ``x`` is given then *add_cyclic* checks if a cyclic point is
        already present by checking all differences between the first and last
        coordinates to be less than ``precision``.
        No point is added if a cyclic point was detected.

        If ``x`` is 1- or 2-dimensional, ``x.shape[-1]`` must equal
        ``data.shape[axis]``, otherwise ``x.shape[axis]`` must equal
        ``data.shape[axis]``.
    y : ndarray, optional
        An n-dimensional array with the values of the y-coordinate, i.e.
        normally the latitudes.
        The cyclic point simply copies the last column. Defaults to None.

        No cyclic point is added if ``y`` is 1-dimensional.
        If ``y`` is 2-dimensional, ``y.shape[-1]`` must equal
        ``data.shape[axis]``, otherwise ``y.shape[axis]`` must equal
        ``data.shape[axis]``.
    axis : int, optional
        Specifies the axis of the arrays to add the cyclic point to,
        i.e. axis with changing x-coordinates. Defaults to the right-most axis.
    cyclic : int or float, optional
        Width of periodic domain (default: 360).
    precision : float, optional
        Maximal difference between first and last x-coordinate to detect
        cyclic point (default: 1e-4).

    Returns
    -------
    cyclic_data
        The data array with a cyclic point added.
    cyclic_x
        The x-coordinate with a cyclic point, only returned if the ``x``
        keyword was supplied.
    cyclic_y
        The y-coordinate with the last column of the cyclic axis duplicated,
        only returned if ``x`` was 2- or n-dimensional and the ``y``
        keyword was supplied.

    Examples
    --------
    Adding a cyclic point to a data array, where the cyclic dimension is
    the right-most dimension.

    >>> import numpy as np
    >>> data = np.ones([5, 6]) * np.arange(6)
    >>> cyclic_data = add_cyclic(data)
    >>> print(cyclic_data)  # doctest: +NORMALIZE_WHITESPACE
    [[0. 1. 2. 3. 4. 5. 0.]
     [0. 1. 2. 3. 4. 5. 0.]
     [0. 1. 2. 3. 4. 5. 0.]
     [0. 1. 2. 3. 4. 5. 0.]
     [0. 1. 2. 3. 4. 5. 0.]]

    Adding a cyclic point to a data array and an associated x-coordinate.

    >>> lons = np.arange(0, 360, 60)
    >>> cyclic_data, cyclic_lons = add_cyclic(data, x=lons)
    >>> print(cyclic_data)  # doctest: +NORMALIZE_WHITESPACE
    [[0. 1. 2. 3. 4. 5. 0.]
     [0. 1. 2. 3. 4. 5. 0.]
     [0. 1. 2. 3. 4. 5. 0.]
     [0. 1. 2. 3. 4. 5. 0.]
     [0. 1. 2. 3. 4. 5. 0.]]
    >>> print(cyclic_lons)
    [  0  60 120 180 240 300 360]

    Adding a cyclic point to a data array and an associated 2-dimensional
    x-coordinate.

    >>> lons = np.arange(0, 360, 60)
    >>> lats = np.arange(-90, 90, 180/5)
    >>> lon2d, lat2d = np.meshgrid(lons, lats)
    >>> cyclic_data, cyclic_lon2d = add_cyclic(data, x=lon2d)
    >>> print(cyclic_data)  # doctest: +NORMALIZE_WHITESPACE
    [[0. 1. 2. 3. 4. 5. 0.]
     [0. 1. 2. 3. 4. 5. 0.]
     [0. 1. 2. 3. 4. 5. 0.]
     [0. 1. 2. 3. 4. 5. 0.]
     [0. 1. 2. 3. 4. 5. 0.]]
    >>> print(cyclic_lon2d)
    [[  0  60 120 180 240 300 360]
     [  0  60 120 180 240 300 360]
     [  0  60 120 180 240 300 360]
     [  0  60 120 180 240 300 360]
     [  0  60 120 180 240 300 360]]

    Adding a cyclic point to a data array and the associated 2-dimensional
    x- and y-coordinates.

    >>> lons = np.arange(0, 360, 60)
    >>> lats = np.arange(-90, 90, 180/5)
    >>> lon2d, lat2d = np.meshgrid(lons, lats)
    >>> cyclic_data, cyclic_lon2d, cyclic_lat2d = add_cyclic(
    ...     data, x=lon2d, y=lat2d)
    >>> print(cyclic_data)  # doctest: +NORMALIZE_WHITESPACE
    [[0. 1. 2. 3. 4. 5. 0.]
     [0. 1. 2. 3. 4. 5. 0.]
     [0. 1. 2. 3. 4. 5. 0.]
     [0. 1. 2. 3. 4. 5. 0.]
     [0. 1. 2. 3. 4. 5. 0.]]
    >>> print(cyclic_lon2d)
    [[  0  60 120 180 240 300 360]
     [  0  60 120 180 240 300 360]
     [  0  60 120 180 240 300 360]
     [  0  60 120 180 240 300 360]
     [  0  60 120 180 240 300 360]]
    >>> print(cyclic_lat2d)
    [[-90. -90. -90. -90. -90. -90. -90.]
     [-54. -54. -54. -54. -54. -54. -54.]
     [-18. -18. -18. -18. -18. -18. -18.]
     [ 18.  18.  18.  18.  18.  18.  18.]
     [ 54.  54.  54.  54.  54.  54.  54.]]

    Not adding a cyclic point if cyclic point detected in x.

    >>> lons = np.arange(0, 361, 72)
    >>> lats = np.arange(-90, 90, 180/5)
    >>> lon2d, lat2d = np.meshgrid(lons, lats)
    >>> cyclic_data, cyclic_lon2d, cyclic_lat2d = add_cyclic(
    ...     data, x=lon2d, y=lat2d)
    >>> print(cyclic_data)  # doctest: +NORMALIZE_WHITESPACE
    [[0. 1. 2. 3. 4. 5.]
     [0. 1. 2. 3. 4. 5.]
     [0. 1. 2. 3. 4. 5.]
     [0. 1. 2. 3. 4. 5.]
     [0. 1. 2. 3. 4. 5.]]
    >>> print(cyclic_lon2d)
    [[  0  72 144 216 288 360]
     [  0  72 144 216 288 360]
     [  0  72 144 216 288 360]
     [  0  72 144 216 288 360]
     [  0  72 144 216 288 360]]
    >>> print(cyclic_lat2d)
    [[-90. -90. -90. -90. -90. -90.]
     [-54. -54. -54. -54. -54. -54.]
     [-18. -18. -18. -18. -18. -18.]
     [ 18.  18.  18.  18.  18.  18.]
     [ 54.  54.  54.  54.  54.  54.]]

    """
    if x is None:
        return _add_cyclic_data(data, axis=axis)
    # if x was given
    if x.ndim > 2:
        xaxis = axis
    else:
        xaxis = -1
    if x.shape[xaxis] != data.shape[axis]:
        estr = (f'x.shape[{xaxis}] does not match the size of the'
                f' corresponding dimension of the data array:'
                f' x.shape[{xaxis}] = {x.shape[xaxis]},'
                f' data.shape[{axis}] = {data.shape[axis]}.')
        raise ValueError(estr)
    if has_cyclic(x, axis=xaxis, cyclic=cyclic, precision=precision):
        if y is None:
            return data, x
        # if y was given
        return data, x, y
    # if not has_cyclic, add cyclic points to data and x
    out_data = _add_cyclic_data(data, axis=axis)
    out_x = _add_cyclic_x(x, axis=xaxis, cyclic=cyclic)
    if y is None:
        return out_data, out_x
    # if y was given
    if y.ndim == 1:
        return out_data, out_x, y
    if y.ndim > 2:
        yaxis = axis
    else:
        yaxis = -1
    if y.shape[yaxis] != data.shape[axis]:
        estr = (f'y.shape[{yaxis}] does not match the size of the'
                f' corresponding dimension of the data array:'
                f' y.shape[{yaxis}] = {y.shape[yaxis]},'
                f' data.shape[{axis}] = {data.shape[axis]}.')
        raise ValueError(estr)
    out_y = _add_cyclic_data(y, axis=yaxis)
    return out_data, out_x, out_y


def clone_ncvmain(widget):
    """
    Duplicate the main ncvue window.

    Parameters
    ----------
    widget : ncvue.ncvMain
        widget of ncvMain class.

    Returns
    -------
    Another ncvue window will be created.

    Examples
    --------
    >>> self.newwin = ttk.Button(
    ...     self.rowwin, text="New Window",
    ...     command=partial(clone_ncvmain, self.master))

    """
    # parent = widget.nametowidget(widget.winfo_parent())
    if widget.name != 'ncvMain':
        print('clone_ncvmain failed. Widget should be ncvMain.')
        print('widget.name is: ', widget.name)
        import sys
        sys.exit()

    root = Toplevel()
    root.name = 'ncvClone'
    root.title("Secondary ncvue window")
    root.geometry('1000x800+150+100')

    root.top = widget.top

    # https://stackoverflow.com/questions/46505982/is-there-a-way-to-clone-a-tkinter-widget
    cls = widget.__class__
    clone = cls(root)
    try:
        for key in widget.configure():
            if key != 'class':
                clone.configure({key: widget.cget(key)})
    except TypeError:
        # in case of CustomTkinter
        cls = ncvue.ncvMain
        clone = cls(root)
        clone.pack(fill=tk.BOTH, expand=1)

    return clone


def format_coord_contour(x, y, ax, xx, yy, zz):
    """
    Formatter function for contour plot including value of nearest array cell.

    Parameters
    ----------
    x, y : float
        Data coordinates of `ax`.
    ax: matplotlib.axes._subplots.AxesSubplot
        Matplotlib axes object for left-hand and right-hand y-axis, resp.
    xx, yy, zz: ndarray
        Numpy arrays with x-values, y-values, and z-values

    Returns
    -------
    String with coordinateds.

    Examples
    --------
    >>> ax = plt.subplot(111)
    >>> ax.pcolormesh(xx, yy, zz)
    >>> ax.format_coord = lambda x, y: format_coord_contour(
    ...     x, y, ax, xx, yy, zz)

    """
    # find closest grid cell
    # https://stackoverflow.com/questions/42577204/show-z-value-at-mouse-pointer-position-in-status-line-with-matplotlibs-pcolorme
    if xx.ndim > 1:
        xarr = xx[0, :]
    else:
        xarr = xx
    # if xx.dtype.type == np.dtype('datetime64').type:
    if np.issubdtype(xx.dtype, np.datetime64):
        xarr = mpld.date2num(xarr)
    if yy.ndim > 1:
        yarr = yy[:, 0]
    else:
        yarr = yy
    # if yy.dtype.type == np.dtype('datetime64').type:
    if np.issubdtype(yy.dtype, np.datetime64):
        yarr = mpld.date2num(yarr)
    if ( (x > xarr.min()) & (x <= xarr.max()) &
         (y > yarr.min()) & (y <= yarr.max()) ):
        col = np.searchsorted(xarr, x) - 1
        row = np.searchsorted(yarr, y) - 1
        xout = xarr[col]
        yout = yarr[row]
        zout = zz[row, col]
    else:
        xout = x
        yout = y
        # if zz.dtype.type == np.dtype('datetime64').type:
        if np.issubdtype(zz.dtype, np.datetime64):
            zout = np.datetime64('NaT')
        else:
            zout = np.nan

    # Special treatment for datetime
    # https://stackoverflow.com/questions/49267011/matplotlib-datetime-from-event-coordinates
    # if xx.dtype.type == np.dtype('datetime64').type:
    if np.issubdtype(xx.dtype, np.datetime64):
        xstr = mpld.num2date(xout).strftime('%Y-%m-%d %H:%M:%S')
    else:
        xstr = '{:.6g}'.format(xout)
    # if yy.dtype.type == np.dtype('datetime64').type:
    if np.issubdtype(yy.dtype, np.datetime64):
        ystr = mpld.num2date(yout).strftime('%Y-%m-%d %H:%M:%S')
    else:
        ystr = '{:.6g}'.format(yout)
    # if zz.dtype.type == np.dtype('datetime64').type:
    if np.issubdtype(zz.dtype, np.datetime64):
        zstr = mpld.num2date(zout).strftime('%Y-%m-%d %H:%M:%S')
    else:
        zstr = '{:.6g}'.format(zout)
    # out = 'x=' + xstr + ', y=' + ystr + ', z=' + zstr
    out = f'x={xstr}, y={ystr}, z={zstr}'
    return out


def format_coord_map(x, y, ax, xx, yy, zz):
    """
    Formatter function for map including value of nearest array cell.

    Parameters
    ----------
    x, y : float
        Data coordinates of `ax`.
    ax: matplotlib.axes._subplots.AxesSubplot
        Matplotlib axes object for left-hand and right-hand y-axis, resp.
    xx, yy, zz: ndarray
        Numpy arrays with x-values, y-values, and z-values

    Returns
    -------
    String with coordinateds.

    Examples
    --------
    >>> ax = plt.subplot(111)
    >>> ax.pcolormesh(xx, yy, zz)
    >>> ax.format_coord = lambda x, y: format_coord_contour(
    ...     x, y, ax, xx, yy, zz)

    """
    # find closest grid cell
    xpp, ypp = ccrs.PlateCarree(central_longitude=0).transform_point(
        x, y, ax.projection)
    x360 = (xpp + 360.) % 360.
    y360 = ypp
    xx360 = (xx + 360.) % 360.
    yy360 = yy
    idx = np.abs((xx360 - x360)**2 + (yy360 - y360)**2).argmin()
    xout = xx.flat[idx]
    yout = yy.flat[idx]
    zout = zz.flat[idx]

    # from cartopy
    lon, lat = ccrs.Geodetic().transform_point(x, y, ax.projection)

    # xstr  = '{:.4g}'.format(xout)
    # ystr  = '{:.4g}'.format(yout)
    xstr  = '{:.4g}'.format(xout)
    ystr  = '{:.4g}'.format(yout)
    zstr  = '{:.6g}'.format(zout)
    ns = 'N' if lat >= 0. else 'S'
    ew = 'E' if lon >= 0. else 'W'
    latstr = u'{:.4f} \u00b0{:s}'.format(abs(lat), ns)
    lonstr = u'{:.4f} \u00b0{:s}'.format(abs(lon), ew)
    # out  = u'x=' + xstr + ', y=' + ystr + ' (' + lonstr + ', ' + latstr + ')'
    # out += ' z=' + zstr
    out = f'x={xstr}, y={ystr} ({lonstr}, {latstr}), z={zstr}'
    return out


def format_coord_scatter(x, y, ax, ax2, xdtype, ydtype, y2dtype):
    """
    Formatter function for scatter plot with left and right axis
    having the same x-axis.

    Parameters
    ----------
    x, y : float
        Data coordinates of `ax2`.
    ax, ax2: matplotlib.axes._subplots.AxesSubplot
        Matplotlib axes object for left-hand and right-hand y-axis, resp.
    xdtype, ydtype, y2dtype: numpy.dtype
        Numpy dtype of data of x-values (xdtype), left-hand side y-values
        (ydtype), and right-hand side y-values (y2dtype)

    Returns
    -------
    String with left-hand side and right hand-side coordinates.

    Examples
    --------
    >>> ax = plt.subplot(111)
    >>> ax2 = ax.twinx()
    >>> ax.plot(xx, yy)
    >>> ax2.plot(xx, yy2)
    >>> ax2.format_coord = lambda x, y: format_coord_scatter(
    ...     x, y, ax, ax2, xx.dtype, yy.dtype, yy2.dtype)

    """
    # convert to display coords
    # https://stackoverflow.com/questions/21583965/matplotlib-cursor-value-with-two-axes
    display_coord = ax2.transData.transform((x, y))
    # convert back to data coords with respect to ax
    inv      = ax.transData.inverted()
    ax_coord = inv.transform(display_coord)

    # Special treatment for datetime
    # https://stackoverflow.com/questions/49267011/matplotlib-datetime-from-event-coordinates
    # if xdtype.type == np.dtype('datetime64').type:
    if np.issubdtype(xdtype, np.datetime64):
        xstr = mpld.num2date(x).strftime('%Y-%m-%d %H:%M:%S')
    else:
        xstr  = '{:.6g}'.format(x)
    # if ydtype.type == np.dtype('datetime64').type:
    if np.issubdtype(ydtype, np.datetime64):
        ystr = mpld.num2date(ax_coord[1]).strftime('%Y-%m-%d %H:%M:%S')
    else:
        ystr  = '{:.6g}'.format(ax_coord[1])
    # if y2dtype.type == np.dtype('datetime64').type:
    if np.issubdtype(y2dtype, np.datetime64):
        y2str = mpld.num2date(y).strftime('%Y-%m-%d %H:%M:%S')
    else:
        y2str = '{:.6g}'.format(y)
    out = f'Left: ({xstr}, {ystr}) Right: ({xstr}, {y2str})'
    return out


def get_slice(dimspins, y):
    """
    Get slice of variable `y` inquiring the spinboxes `dimspins`.

    Parameters
    ----------
    dimspins : list
        List of tk.Spinbox widgets of dimensions
    y : ndarray or netCDF4._netCDF4.Variable
        Input array or netcdf variable

    Returns
    -------
    ndarray
        Slice of `y` chosen with spinboxes.

    Examples
    --------
    >>> gy, vy = vardim2var(y, self.groups)
    >>> yy = selvar(self, vy)
    >>> miss = get_miss(self, yy)
    >>> yy = get_slice(self.yd, yy).squeeze()
    >>> yy = set_miss(miss, yy)

    """
    isxarray = False
    if ihavex:
        if isinstance(y, xr.DataArray):
            isxarray = True
    methods = ['all']
    methods.extend(DIMMETHODS)
    dd = []
    ss = []
    for i in range(y.ndim):
        dim = dimspins[i].get()
        if dim in methods:
            s = slice(0, y.shape[i])
        else:
            idim = int(dim)
            s = slice(idim, idim + 1)
        dd.append(dim)
        ss.append(s)
    if len(ss) > 0:
        imeth = list_intersection(dd, DIMMETHODS)
        if len(imeth) > 0:
            if isxarray:
                sargs = dict(zip(y.dims, ss))
                yout = y.isel(**sargs).to_numpy()
            else:
                yout = y[tuple(ss)]
            ii = [ i for i, d in enumerate(dd) if d in imeth ]
            ii.reverse()  # last axis first
            for i in ii:
                if dd[i] == 'mean':
                    yout = np.ma.mean(yout, axis=i)
                elif dd[i] == 'std':
                    yout = np.ma.std(yout, axis=i)
                elif dd[i] == 'min':
                    yout = np.ma.min(yout, axis=i)
                elif dd[i] == 'max':
                    yout = np.ma.max(yout, axis=i)
                elif dd[i] == 'ptp':
                    yout = np.ma.ptp(yout, axis=i)
                elif dd[i] == 'sum':
                    yout = np.ma.sum(yout, axis=i)
                elif dd[i] == 'median':
                    yout = np.ma.median(yout, axis=i)
                elif dd[i] == 'var':
                    yout = np.ma.var(yout, axis=i)
            return yout
        else:
            if isxarray:
                sargs = dict(zip(y.dims, ss))
                yout = y.isel(**sargs).to_numpy()
            else:
                yout = y[tuple(ss)]
            return yout
    else:
        return np.array([], dtype=y.dtype)


def get_standard_name(ivar):
    """
    Get standard_name, long_name, or name attribute of netcdf variable

    Parameters
    ----------
    ivar : netCDF4._netCDF4.Variable
        netcdf variable

    Returns
    -------
    string
        standard_name, long_name, or name of variable `ivar`

    Examples
    --------
    >>> name = get_standard_name(self.fi['Tair])

    """
    if hasattr(ivar, 'standard_name'):
        return ivar.standard_name
    elif hasattr(ivar, 'long_name'):
        return ivar.long_name
    else:
        return ivar.name


def get_units(ivar):
    """
    Get units attribute of netcdf variable

    Parameters
    ----------
    ivar : netCDF4._netCDF4.Variable
        netcdf variable

    Returns
    -------
    string
        units or empty string

    Examples
    --------
    >>> units = get_units(self.fi['Tair])

    """
    if hasattr(ivar, 'units'):
        return ivar.units
    else:
        return ''


def list_intersection(lst1, lst2):
    """
    Intersection of two lists.

    From:
    https://stackoverflow.com/questions/3697432/how-to-find-list-intersection
    Using list comprehension for small lists and set() method with builtin
    intersection for longer lists.

    Parameters
    ----------
    lst1, lst2 : list
        Python lists

    Returns
    -------
    list
        List with common elements in both input lists.

    Examples
    --------
    >>> lst1 = [ 4, 9, 1, 17, 11, 26, 28, 28, 26, 66, 91]
    >>> lst2 = [9, 9, 74, 21, 45, 11, 63]
    >>> print(Intersection(lst1, lst2))
    [9, 11]

    """
    if (len(lst1) > 10) or (len(lst2) > 10):
        return list(set(lst1).intersection(lst2))
    else:
        return [ ll for ll in lst1 if ll in lst2 ]


def selvar(self, var):
    """
    Extract variable from correct file.

    Parameters
    ----------
    self : class
        ncvue Tk class having fi and groups
    var : string
        Variable name including group or file identifier,
        such as 'Qle', 'group1/Qle', or 'file0001/Qle'

    Returns
    -------
    netcdf variable
        Pointer to variable in file

    Examples
    --------
    selvar(self, var)

    """
    if len(self.fi) == 0:
        return
    elif self.usex:
        return self.fi[var]
    elif len(self.fi) == 1:
        fil = self.fi[0]
        return fil[var]
    else:
        if '/' in var:
            f = var[0:var.find('/')].rstrip()
            idxf = list(self.groups).index(f)
            vv = var[var.find('/') + 1:].rstrip()
            fil = self.fi[idxf]
            return fil[vv]
        else:
            fil = self.fi[0]
            return fil[var]


def set_axis_label(ncvar):
    """
    Set label plotting axis from name and unit of given variable
    `ncvar`.

    Parameters
    ----------
    ncvar : netCDF4._netCDF4.Variable
        netcdf variables class

    Returns
    -------
    str
        Label string: name (unit)

    Examples
    --------
    >>> ylab = set_axis_label(fi['w_soil'])

    """
    try:
        lab = ncvar.long_name
    except AttributeError:
        try:
            lab = ncvar.standard_name
        except AttributeError:
            lab = ncvar.name
    try:
        lab += ' (' + ncvar.units + ')'
    except AttributeError:
        pass
    return lab


def set_miss(miss, x):
    """
    Set `x` to NaN or NaT for all values in miss.

    Parameters
    ----------
    miss : iterable
        values which shall be set to np.nan or np.datetime64('NaT') in `x`
    x : ndarray
        numpy array

    Returns
    -------
    ndarray
        `x` with all values set np.nan or np.datetime64('NaT')
        that are equal to any value in miss.

    Examples
    --------
    >>> x = fi['time']
    >>> miss = get_miss(self, x)
    >>> x = set_miss(miss, x)

    """
    if np.issubdtype(x.dtype, np.datetime64):
        default = np.datetime64('NaT')
    else:
        default = np.nan
    for mm in miss:
        x = np.where(x == mm, default, x)
    return x


def spinbox_values(ndim):
    """
    Tuple for Spinbox values with 'all' before range(`ndim`) and
    'mean', 'std', etc. after range(`ndim`) if `ndim`>1,
    otherwise single entry (0,).

    Parameters
    ----------
    ndim : int
        Size of dimension.

    Returns
    -------
    tuple
        (('all',) + tuple(range(ndim)) + ('mean','std',...)) if ndim > 1
        (0,) else

    Examples
    --------
    >>> self.xd0.config(values=spinbox_values(xx.shape[0]))

    """
    if ndim > 1:
        return (('all',) + tuple(range(ndim)) + DIMMETHODS)
    else:
        return (0,)


def vardim2var(vardim, groups=[]):
    """
    Extract group index and variable name from 'group/variable (dim1=ndim1,)'.

    Parameters
    ----------
    vardim : string
        Variable name with dimensions, such as 'latitude (lat=32,lon=64)'
        or with group 'file1/latitude (lat=32,lon=64)'

    Returns
    -------
    int, string
        group index, variable name

    Examples
    --------
    >>> vardim2var('latitude (lat=32,lon=64)')
    0 latitude

    """
    var = vardim[0:vardim.rfind('(')].rstrip()
    if '/' in var:
        g = var[0:var.find('/')].rstrip()
        ig = list(groups).index(g)
    else:
        ig = 0
    return ig, var


def xzip_dim_name_length(ncvar):
    """
    Combines dimension names and length of xarray variable in list of strings.

    Parameters
    ----------
    ncvar : xarray.DataArray
        xarray variable

    Returns
    -------
    list
        List of dimension name and length in the form 'dim=len'.

    Examples
    --------
    >>> import xarray as xr
    >>> ifile = 'test.nc'
    >>> fi = xr.open_dataset(ifile)
    >>> w_soil = fi['w_soil']
    >>> print(zip_dim_name_length(w_soil))
    ('ntime=17520', 'nsoil=30')

    """
    out = []
    for i in range(ncvar.ndim):
        out.append(f'{ncvar.dims[i]}={ncvar.shape[i]}')
    return out


def zip_dim_name_length(ncvar):
    """
    Combines dimension names and length of netcdf variable in list of strings.

    Parameters
    ----------
    ncvar : netCDF4._netCDF4.Variable
        netcdf variables class

    Returns
    -------
    list
        List of dimension name and length in the form 'dim=len'.

    Examples
    --------
    >>> import netCDF4 as nc
    >>> ifile = 'test.nc'
    >>> fi = nc.Dataset(ifile, 'r')
    >>> w_soil = fi['w_soil']
    >>> print(zip_dim_name_length(w_soil))
    ('ntime=17520', 'nsoil=30')

    """
    out = []
    for i in range(ncvar.ndim):
        out.append(f'{ncvar.dimensions[i]}={ncvar.shape[i]}')
    return out



#!/usr/bin/env python
"""
Common methods for panels of ncvue.

Methods normally belong to a class and be called like self.method(args).
We have several panels in ncvue that would have the same methods such as
getting slices from arrays. These are hence gathered here, all start with
the first argument self, which is the class instance, and are called like
method(self, args).

The methods could also be put onto a common class (based on ttk.Frame) on
which ncvContour, etc. would then be based.

This module was written by Matthias Cuntz while at Institut National de
Recherche pour l'Agriculture, l'Alimentation et l'Environnement (INRAE), Nancy,
France.

:copyright: Copyright 2020-2021 Matthias Cuntz - mc (at) macu (dot) de
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following methods are provided:

.. autosummary::
   analyse_netcdf
   get_miss
   get_slice_miss
   set_dim_lat
   set_dim_lon
   set_dim_var
   set_dim_x
   set_dim_y
   set_dim_y2
   set_dim_z

History
   * Written Nov-Dec 2020 by Matthias Cuntz (mc (at) macu (dot) de)
   * Slice arrays with slice function rather than numpy.take,
     Dec 2020, Matthias Cuntz
   * Moved individual get_slice? methods for x, y, y2, z as general get_slice
     function to ncvutils, Dec 2020, Matthias Cuntz
   * Added convenience method get_slice_miss, Dec 2020, Matthias Cuntz
   * set_dim_lon, set_dim_lat, set_dim_var for Map panel,
     Jan 2021, Matthias Cuntz
   * Set latdim, londim to all on map var, determined in ncvMain,
     Jan 2021, Matthias Cuntz
   * Catch non numpy.dtype in set_miss, Jan 2021, Matthias Cuntz
   * Catch variables that have only one string or similar,
     Jan 2021, Matthias Cuntz
   * Added tooltip to dimensions, Jan 2021, Matthias Cuntz
   * Added analyse_netcdf from ncvmain, Jan 2021, Matthias Cuntz
   * Use dlblval instead of dlbl to set dimension labels,
     Jan 2021, Matthias Cuntz
   * Make self.time numpy's datetime64[ms] format, May 2021, Matthias Cuntz
   * Add numpy's datetime64[ms] to missing values, May 2021, Matthias Cuntz
   * Removed bug in detection of lon/lat (analyse_netcdf),
     Oct 2021, Matthias Cuntz
   * Identify lon/lat also by axis attributes x/y or X/Y (analyse_netcdf),
     Oct 2021, Matthias Cuntz
   * Do not default the unlimited dimension to 'all' if no lon/lat were found,
     (get_dim_var) Oct 2021, Matthias Cuntz
   * Address fi.variables[name] directly by fi[name], Jan 2024, Matthias Cuntz
   * Allow groups in netcdf files, Jan 2024, Matthias Cuntz
   * Squeeze output in get_slice_miss only if more than 1 dim,
     Jan 2024, Matthias Cuntz
   * Allow multiple netcdf files, Jan 2024, Matthias Cuntz
   * Rename analyse_netcdf to analyse_netcdf_xarray_ncvue,
     add new analyse_netcdf, Feb 2025, Matthias Cuntz
   * Use new get_standard_name and get_units from ncvutils,
     Feb 2025, Matthias Cuntz
   * Add analyse_netcdf_xarray, Feb 2025, Matthias Cuntz

"""
import tkinter as tk
import numpy as np
from .ncvutils import DIMMETHODS, get_slice, set_miss, spinbox_values
from .ncvutils import vardim2var, selvar
from .ncvutils import xzip_dim_name_length, zip_dim_name_length
from .ncvutils import get_standard_name, get_units
import netCDF4 as nc
# nc.default_fillvals but with keys as variables['var'].dtype
nctypes = [ np.dtype(i) for i in nc.default_fillvals ]
ncfill  = dict(zip(nctypes, list(nc.default_fillvals.values())))
ncfill.update({np.dtype('O'): np.nan})
ncfill.update({np.dtype('<M8[ms]'): np.datetime64('NaT')})
ncfill.update({np.dtype('<M8[ns]'): np.datetime64('NaT')})


__all__ = ['analyse_netcdf',
           'get_miss', 'get_slice_miss',
           'set_dim_lat', 'set_dim_lon', 'set_dim_var',
           'set_dim_x', 'set_dim_y', 'set_dim_y2', 'set_dim_z']


#
# Analyse netcdf file
#

def analyse_netcdf(self):
    """
    Call analyse_netcdf_xarray or analyse_netcdf_ncvue depending on self.usex

    Parameters
    ----------
    self : class
        ncvue class

    Returns
    -------
    Set variables:
        self.dunlim,
        self.time, self.tname, self.tvar, self.dtime,
        self.cols,
        self.latvar, self.lonvar, self.latdim, self.londim

    Examples
    --------
    >>> analyse_netcdf(self)

    """
    if self.usex:
        analyse_netcdf_xarray(self)
    else:
        analyse_netcdf_ncvue(self)


def analyse_netcdf_ncvue(self):
    """
    Analyse netcdf file(s) for the unlimited dimension, calculating datetime,
    variables, latitudes/longitudes variables and dimensions.

    Parameters
    ----------
    self : class
        ncvue class

    Returns
    -------
    Set variables:
        self.dunlim,
        self.time, self.tname, self.tvar, self.dtime,
        self.cols,
        self.latvar, self.lonvar, self.latdim, self.londim

    Examples
    --------
    >>> analyse_netcdf(self)

    """
    import datetime as dt
    try:
        import cftime as cf
    except ModuleNotFoundError:
        import netCDF4 as cf
    #
    ngroups = len(self.groups)
    for ig in range(max(ngroups, 1)):
        if len(self.fi) == 1:
            if ngroups > 0:
                ffi = self.fi[0]
                fi = ffi[self.groups[ig]]
                gname = self.groups[ig] + '/'
            else:
                fi = self.fi[ig]
                gname = ''
        else:
            fi = self.fi[ig]
            gname = self.groups[ig] + '/'
        #
        # search unlimited dimension
        self.dunlim.append('')
        for dd in fi.dimensions:
            if fi.dimensions[dd].isunlimited():
                self.dunlim[ig] = dd
                break
        #
        # search for time variable and make datetime variable
        self.time.append(None)
        self.tname.append('')
        self.tvar.append('')
        self.dtime.append(None)
        for vv in fi.variables:
            isunlim = False
            if self.dunlim[ig]:
                if vv.lower() == fi.dimensions[self.dunlim[ig]].name.lower():
                    isunlim = True
            if ( isunlim or vv.lower().startswith('time_') or
                 (vv.lower() == 'time') or (vv.lower() == 'datetime') or
                 (vv.lower() == 'date') ):
                self.tvar[ig] = gname + vv
                if vv.lower() == 'datetime':
                    self.tname[ig] = gname + 'date'
                else:
                    self.tname[ig] = gname + 'datetime'
                try:
                    ivar = selvar(self, self.tvar[ig])
                    tunit = ivar.units
                except AttributeError:
                    tunit = ''
                # assure 01, etc. if values < 1000, 10, 10 in year, month, day
                if tunit.find('since') > 0:
                    tt = tunit.split()
                    dd = tt[2].split('-')
                    tt[2] = (f'{int(dd[0]):04d}-{int(dd[1]):02d}-'
                             f'{int(dd[2]):02d}')
                    tunit = ' '.join(tt)
                try:
                    ivar = selvar(self, self.tvar[ig])
                    tcal = ivar.calendar
                except AttributeError:
                    tcal = 'standard'
                ivar = selvar(self, self.tvar[ig])
                time = ivar[:]
                # time dimension "day as %Y%m%d.%f" from cdo.
                if ' as ' in tunit:
                    itunit = tunit.split()[2]
                    dtime = []
                    for tt in time:
                        stt = str(tt).split('.')
                        sstt = ('00' + stt[0])[-8:] + '.' + stt[1]
                        dtime.append(dt.datetime.strptime(sstt, itunit))
                    ntime = cf.date2num(dtime,
                                        'days since 0001-01-01 00:00:00')
                    self.dtime[ig] = cf.num2date(
                        ntime,
                        'days since 0001-01-01 00:00:00')
                else:
                    try:
                        self.dtime[ig] = cf.num2date(time, tunit,
                                                     calendar=tcal)
                    except ValueError:
                        self.dtime[ig] = None
                if self.dtime[ig] is not None:
                    ntime = len(self.dtime[ig])
                    if (tcal == '360_day'):
                        ndays = [360.] * ntime
                    elif (tcal == '365_day'):
                        ndays = [365.] * ntime
                    elif (tcal == 'noleap'):
                        ndays = [365.] * ntime
                    elif (tcal == '366_day'):
                        ndays = [366.] * ntime
                    elif (tcal == 'all_leap'):
                        ndays = [366.] * ntime
                    else:
                        ndays = [ 365. +
                                  float((((t.year % 4) == 0) &
                                         ((t.year % 100) != 0)) |
                                        ((t.year % 400) == 0))
                                  for t in self.dtime[ig] ]
                    self.dtime[ig] = np.array([
                        t.year +
                        (t.dayofyr - 1 + t.hour / 24. +
                         t.minute / 1440 + t.second / 86400.) / ndays[i]
                        for i, t in enumerate(self.dtime[ig]) ])
                # make datetime variable
                if self.time[ig] is None:
                    try:
                        ttime = cf.num2date(
                            time, tunit, calendar=tcal,
                            only_use_cftime_datetimes=False,
                            only_use_python_datetimes=True)
                        self.time[ig] = np.array([ dd.isoformat()
                                                   for dd in ttime ],
                                                 dtype='datetime64[ms]')
                    except:
                        self.time[ig] = None
                if self.time[ig] is None:
                    try:
                        ttime = cf.num2date(time, tunit,
                                            calendar=tcal)
                        self.time[ig] = np.array([ dd.isoformat()
                                                   for dd in ttime ],
                                                 dtype='datetime64[ms]')
                    except:
                        self.time[ig] = None
                if self.time[ig] is None:
                    # if not possible use decimal year
                    self.time[ig] = self.dtime[ig]
                if self.time[ig] is None:
                    # could not interpret time at all,
                    # e.g. if units = "months since ..."
                    self.time[ig] = time
                    self.dtime[ig] = time
                break
        #
        # construct list of variable names with dimensions
        if self.time[ig] is not None:
            ivar = selvar(self, self.tvar[ig])
            addt = [self.tname[ig] + ' ' +
                    str(tuple(zip_dim_name_length(ivar)))]
            self.cols += addt
        ivars = []
        for vv in fi.variables:
            vname = gname + vv
            ss = tuple(zip_dim_name_length(fi[vv]))
            self.maxdim = max(self.maxdim, len(ss))
            ivars.append((vname, ss, len(ss)))
        self.cols += sorted([ vv[0] + ' ' + str(vv[1])
                              for vv in ivars ])
        #
        # search for lat/lon variables
        self.latvar.append('')
        self.lonvar.append('')
        # first sweep: *name must be "latitude" and
        #              units must be "degrees_north"
        if not self.latvar[ig]:
            for vv in fi.variables:
                sname = get_standard_name(fi[vv])
                if sname.lower() == 'latitude':
                    sunit = get_units(fi[vv])
                    if sunit.lower() == 'degrees_north':
                        self.latvar[ig] = gname + vv
        if not self.lonvar[ig]:
            for vv in fi.variables:
                sname = get_standard_name(fi[vv])
                if sname.lower() == 'longitude':
                    sunit = get_units(fi[vv])
                    if sunit.lower() == 'degrees_east':
                        self.lonvar[ig] = gname + vv
        # second sweep: name must start with lat and
        #               units must be "degrees_north"
        if not self.latvar[ig]:
            for vv in fi.variables:
                sname = fi[vv].name
                if sname[0:3].lower() == 'lat':
                    sunit = get_units(fi[vv])
                    if sunit.lower() == 'degrees_north':
                        self.latvar[ig] = gname + vv
        if not self.lonvar[ig]:
            for vv in fi.variables:
                sname = fi[vv].name
                if sname[0:3].lower() == 'lon':
                    sunit = get_units(fi[vv])
                    if sunit.lower() == 'degrees_east':
                        self.lonvar[ig] = gname + vv
        # third sweep: name must contain lat and
        #              units must be "degrees_north"
        if not self.latvar[ig]:
            for vv in fi.variables:
                sname = fi[vv].name
                sname = sname.lower()
                if sname.find('lat') >= 0:
                    sunit = get_units(fi[vv])
                    if sunit.lower() == 'degrees_north':
                        self.latvar[ig] = gname + vv
        if not self.lonvar[ig]:
            for vv in fi.variables:
                sname = fi[vv].name
                sname = sname.lower()
                if sname.find('lon') >= 0:
                    sunit = get_units(fi[vv])
                    if sunit.lower() == 'degrees_east':
                        self.lonvar[ig] = gname + vv
        # fourth sweep: axis is 'Y' or 'y'
        if not self.latvar[ig]:
            for vv in fi.variables:
                try:
                    saxis = fi[vv].axis
                except AttributeError:
                    saxis = ''
                if saxis.lower() == 'y':
                    self.latvar[ig] = gname + vv
        if not self.lonvar[ig]:
            for vv in fi.variables:
                try:
                    saxis = fi[vv].axis
                except AttributeError:
                    saxis = ''
                if saxis.lower() == 'x':
                    self.lonvar[ig] = gname + vv
        # fifth sweep: same as first but units can be simply "degrees"
        if not self.latvar[ig]:
            for vv in fi.variables:
                sname = get_standard_name(fi[vv])
                if sname.lower() == 'latitude':
                    sunit = get_units(fi[vv])
                    if sunit.lower() == 'degrees':
                        self.latvar[ig] = gname + vv
        if not self.lonvar[ig]:
            for vv in fi.variables:
                sname = get_standard_name(fi[vv])
                if sname.lower() == 'longitude':
                    sunit = get_units(fi[vv])
                    if sunit.lower() == 'degrees':
                        self.lonvar[ig] = gname + vv
        # sixth sweep: same as second but units can be simply "degrees"
        if not self.latvar[ig]:
            for vv in fi.variables:
                sname = fi[vv].name
                if sname[0:3].lower() == 'lat':
                    sunit = get_units(fi[vv])
                    if sunit.lower() == 'degrees':
                        self.latvar[ig] = gname + vv
        if not self.lonvar[ig]:
            for vv in fi.variables:
                sname = fi[vv].name
                if sname[0:3].lower() == 'lon':
                    sunit = get_units(fi[vv])
                    if sunit.lower() == 'degrees':
                        self.lonvar[ig] = gname + vv
        # seventh sweep: same as third but units can be simply "degrees"
        if not self.latvar[ig]:
            for vv in fi.variables:
                sname = fi[vv].name
                sname = sname.lower()
                if sname.find('lat') >= 0:
                    sunit = get_units(fi[vv])
                    if sunit.lower() == 'degrees':
                        self.latvar[ig] = gname + vv
        if not self.lonvar[ig]:
            for vv in fi.variables:
                sname = fi[vv].name
                sname = sname.lower()
                if sname.find('lon') >= 0:
                    sunit = get_units(fi[vv])
                    if sunit.lower() == 'degrees':
                        self.lonvar[ig] = gname + vv
        #
        # determine lat/lon dimensions
        self.latdim.append('')
        self.londim.append('')
        if self.latvar[ig]:
            ivar = selvar(self, self.latvar[ig])
            latshape = ivar.shape
            if (len(latshape) < 1) or (len(latshape) > 2):
                print('Something went wrong determining lat/lon:'
                      ' latitude variable is not 1D or 2D.\n'
                      'latitude variable with dimensions:',
                      self.latvar[ig], ivar.dimensions)
                self.latvar[ig] = ''
            else:
                self.latdim[ig] = ivar.dimensions[0]
        if self.lonvar[ig]:
            ivar = selvar(self, self.lonvar[ig])
            lonshape = ivar.shape
            if len(lonshape) == 1:
                self.londim[ig] = ivar.dimensions[0]
            elif len(lonshape) == 2:
                self.londim[ig] = ivar.dimensions[1]
            else:
                print('Something went wrong determining lat/lon:'
                      ' longitude variable is not 1D or 2D.\n'
                      'longitude variable with dimensions:',
                      self.lonvar[ig], ivar.dimensions)
                self.lonvar[ig] = ''
        #
        # add units to lat/lon name
        if self.latvar[ig]:
            ivar = selvar(self, self.latvar[ig])
            idim = tuple(zip_dim_name_length(ivar))
            self.latvar[ig] = self.latvar[ig] + ' ' + str(idim)
        if self.lonvar[ig]:
            ivar = selvar(self, self.lonvar[ig])
            idim = tuple(zip_dim_name_length(ivar))
            self.lonvar[ig] = self.lonvar[ig] + ' ' + str(idim)


def analyse_netcdf_xarray(self):
    """
    Analyse netcdf file(s) opened with xarray for time (= unlimited),
    variables, latitudes/longitudes variables and dimensions.

    Parameters
    ----------
    self : class
        ncvue class

    Returns
    -------
    Set variables:
        self.dunlim,
        self.time, self.tname, self.tvar, self.dtime,
        self.cols,
        self.latvar, self.lonvar, self.latdim, self.londim

    Examples
    --------
    >>> analyse_netcdf_xarray(self)

    """
    import xarray as xr

    #
    # search time
    self.dunlim = None
    self.time = None
    self.tname = ''
    self.tvar = None
    self.dtime = None
    for cc in self.fi.coords:
        if np.issubdtype(self.fi.coords[cc].dtype, np.datetime64):
            self.dunlim = cc
            break
    if self.dunlim is None:
        if 'time' in self.fi.coords:
            self.dunlim = 'time'
    if self.dunlim is not None:
        self.time = self.fi.coords[self.dunlim]
        self.tname = self.dunlim
        self.tvar = self.dunlim
        self.dtime = self.time

    #
    # construct list of variable names with dimensions
    if self.time is not None:
        addt = [self.tname + ' ' +
                str(tuple(xzip_dim_name_length(self.fi[self.tvar])))]
        self.cols += addt
    ivars = []
    for vv in self.fi.variables:
        if vv != self.tname:
            ss = tuple(xzip_dim_name_length(self.fi[vv]))
            ndim = self.fi[vv].ndim
            self.maxdim = max(self.maxdim, ndim)
            ivars.append((vv, ss, ndim))
    self.cols += sorted([ vv[0] + ' ' + str(vv[1])
                          for vv in ivars ])

    #
    # search for lat/lon variables
    # first sweep: units must be "degrees_north" and "degrees_east"
    self.latvar = ''
    self.lonvar = ''
    for cc in self.fi.coords:
        ccoord = self.fi.coords[cc]
        cunits = get_units(ccoord)
        if cunits == 'degrees_north':
            self.latvar = cc
        if cunits == 'degrees_east':
            self.lonvar = cc
    # second sweep: axis is 'x', 'y', or 'X', 'Y'
    if not self.latvar:
        for cc in self.fi.coords:
            if cc.lower() == 'y':
                self.latvar = cc
    if not self.lonvar:
        for cc in self.fi.coords:
            if cc.lower() == 'x':
                self.lonvar = cc

    #
    # determine lat/lon dimensions
    self.latdim = ''
    self.londim = ''

    #
    # add units to lat/lon name
    if self.latvar:
        idim = tuple(xzip_dim_name_length(self.fi[self.latvar]))
        self.latvar = self.latvar + ' ' + str(idim)
    if self.lonvar:
        idim = tuple(xzip_dim_name_length(self.fi[self.lonvar]))
        self.lonvar = self.lonvar + ' ' + str(idim)


#
# Get list of missing values
#

def get_miss(self, x):
    """
    Get list of missing values, i.e. self.miss, x._FillValue,
    x.missing_value, and from netcdf4.default_fillvals.

    Parameters
    ----------
    self : class
        ncvue class
    x : netCDF4._netCDF4.Variable
        netcdf variable

    Returns
    -------
    list
        List with missing values self.miss, x._FillValue,
        x.missing_value if present,  and from netcdf4.default_fillvals

    Examples
    --------
    >>> x = fi['time']
    >>> miss = get_miss(self, x)

    """
    try:
        out = [ncfill[x.dtype]]
    except KeyError:
        out = []
    try:
        if x.dtype != np.dtype('<M8[ms]'):
            out += [self.miss]
    except AttributeError:
        pass
    try:
        out += [x._FillValue]
    except AttributeError:
        pass
    try:
        out += [x.missing_value]
    except AttributeError:
        pass
    return out


#
# Get slice and set missing value
#

def get_slice_miss(self, dimspins, x):
    """
    Convenience method to get list of missing values (get_miss),
    choose slice of array (get_slice), and set missing values to
    NaN or np.datetime64('NaT') in slice (set_miss).

    Parameters
    ----------
    self : class
        ncvue class
    dimspins : list
        List of tk.Spinbox widgets of dimensions
    x : netCDF4._netCDF4.Variable
        netcdf variable

    Returns
    -------
    ndarray
        Extracted array slice with missing values set to
        np.NaN or np.datetime64('NaT')

    Examples
    --------
    >>> x = fi['time']
    >>> xx = get_slice_miss(self, x)

    """
    miss = get_miss(self, x)
    xx = get_slice(dimspins, x)
    if xx.ndim > 1:
        xx = xx.squeeze()
    xx = set_miss(miss, xx)
    # catch variables that have only one string or similar
    try:
        sx = xx.shape[0]
    except IndexError:
        xx = np.array([np.nan])
    return xx


#
# Set dimensions
#

# set_dim_lat/lon/var could also be methods of ncvMap.
# Leave them here for their concordance with set_dim_x/y/z

def set_dim_lat(self):
    """
    Set spinboxes of latitude-dimensions.

    Set labels and value lists. Select 'all' for all dimension.

    Parameters
    ----------
    self : class
        ncvue class

    Returns
    -------
    None
        Labels and values of spinboxes of latitude-dimensions set.

    Examples
    --------
    >>> set_dim_lat(self)

    """
    # reset dimensions
    for i in range(self.maxdim):
        self.latd[i].config(values=(0,), width=1, state=tk.DISABLED)
        self.latdlblval[i].set(str(i))
        self.latdtip[i].set("")
        self.latdframe[i].pack(side='left')
    lat = self.lat.get()
    if lat != '':
        # set real dimensions
        gl, vl = vardim2var(lat, self.groups)
        if self.usex:
            if vl == self.tname:
                vl = self.tvar
        else:
            if vl == self.tname[gl]:
                vl = self.tvar[gl]
        ll = selvar(self, vl)
        for i in range(ll.ndim):
            ww = max(4, int(np.ceil(np.log10(ll.shape[i]))))
            self.latd[i].config(values=spinbox_values(ll.shape[i]), width=ww,
                                state=tk.NORMAL)
            if (ll.shape[i] > 1):
                self.latdval[i].set('all')
            else:
                self.latdval[i].set('0')
            if self.usex:
                self.latdlblval[i].set(str(ll.dims[i]))
            else:
                self.latdlblval[i].set(str(ll.dimensions[i]))
            self.latdframe[i].pack(side='left')
            if ll.shape[i] > 1:
                tstr  = "Specific dimension value: 0-{:d}\n".format(
                    ll.shape[i] - 1)
                tstr += "or arithmetic operation on axis:\n"
                tstr += "  " + ", ".join(DIMMETHODS)
            else:
                tstr = "Single dimension: 0"
            self.latdtip[i].set(tstr)


def set_dim_lon(self):
    """
    Set spinboxes of longitude-dimensions.

    Set labels and value lists. Select 'all' for all dimension.

    Parameters
    ----------
    self : class
        ncvue class

    Returns
    -------
    None
        Labels and values of spinboxes of longitude-dimensions set.

    Examples
    --------
    >>> set_dim_lon(self)

    """
    # reset dimensions
    for i in range(self.maxdim):
        self.lond[i].config(values=(0,), width=1, state=tk.DISABLED)
        self.londlblval[i].set(str(i))
        self.londtip[i].set("")
        self.londframe[i].pack(side='left')
    lon = self.lon.get()
    if lon != '':
        # set real dimensions
        gl, vl = vardim2var(lon, self.groups)
        if self.usex:
            if vl == self.tname:
                vl = self.tvar
        else:
            if vl == self.tname[gl]:
                vl = self.tvar[gl]
        ll = selvar(self, vl)
        for i in range(ll.ndim):
            ww = max(4, int(np.ceil(np.log10(ll.shape[i]))))
            self.lond[i].config(values=spinbox_values(ll.shape[i]), width=ww,
                                state=tk.NORMAL)
            if (ll.shape[i] > 1):
                self.londval[i].set('all')
            else:
                self.londval[i].set('0')
            if self.usex:
                self.londlblval[i].set(str(ll.dims[i]))
            else:
                self.londlblval[i].set(str(ll.dimensions[i]))
            self.londframe[i].pack(side='left')
            if ll.shape[i] > 1:
                tstr  = "Specific dimension value: 0-{:d}\n".format(
                    ll.shape[i] - 1)
                tstr += "or arithmetic operation on axis:\n"
                tstr += "  " + ", ".join(DIMMETHODS)
            else:
                tstr = "Single dimension: 0"
            self.londtip[i].set(tstr)


def set_dim_var(self):
    """
    Set spinboxes of varable-dimensions in Map panel.

    Set labels and value lists, including 'all' to select all entries,
    as well as 'mean', 'std', etc. for common operations on the axis.

    Select 'all' for the first two limited dimensions and select
    0 for all other dimensions.

    Parameters
    ----------
    self : class
        ncvue class

    Returns
    -------
    None
        Labels and values of spinboxes of variable-dimensions set.

    Examples
    --------
    >>> set_dim_var(self)

    """
    # reset dimensions
    for i in range(self.maxdim):
        self.vd[i].config(values=(0,), width=1, state=tk.DISABLED)
        self.vdlblval[i].set(str(i))
        self.vdtip[i].set("")
        self.vdframe[i].pack(side='left')
    v = self.v.get()
    if v != '':
        # set real dimensions
        gz, vz = vardim2var(v, self.groups)
        if self.usex:
            if vz == self.tname:
                vz = self.tvar
        else:
            if vz == self.tname[gz]:
                vz = self.tvar[gz]
        vv = selvar(self, vz)
        if self.usex:
            latdim = self.latdim
            londim = self.londim
            dunlim = self.dunlim
            dims = vv.dims
        else:
            latdim = self.latdim[gz]
            londim = self.londim[gz]
            dunlim = self.dunlim[gz]
            dims = vv.dimensions
        nall = 0
        if latdim:
            if latdim in dims:
                i = dims.index(latdim)
                ww = max(5, int(np.ceil(np.log10(vv.shape[i]))))  # 5~median
                self.vd[i].config(values=spinbox_values(vv.shape[i]), width=ww,
                                  state=tk.NORMAL)
                nall += 1
                self.vdval[i].set('all')
                self.vdlblval[i].set(str(dims[i]))
                self.vdframe[i].pack(side='left')
                if vv.shape[i] > 1:
                    tstr  = "Specific dimension value: 0-{:d}\n".format(
                        vv.shape[i] - 1)
                    tstr += "or arithmetic operation on axis:\n"
                    tstr += "  " + ", ".join(DIMMETHODS)
                else:
                    tstr = "Single dimension: 0"
                self.vdtip[i].set(tstr)
        if londim:
            if londim in dims:
                i = vv.dimensions.index(londim)
                ww = max(5, int(np.ceil(np.log10(vv.shape[i]))))  # 5~median
                self.vd[i].config(values=spinbox_values(vv.shape[i]), width=ww,
                                  state=tk.NORMAL)
                nall += 1
                self.vdval[i].set('all')
                self.vdlblval[i].set(str(dims[i]))
                self.vdframe[i].pack(side='left')
                if vv.shape[i] > 1:
                    tstr  = "Specific dimension value: 0-{:d}\n".format(
                        vv.shape[i] - 1)
                    tstr += "or arithmetic operation on axis:\n"
                    tstr += "  " + ", ".join(DIMMETHODS)
                else:
                    tstr = "Single dimension: 0"
                self.vdtip[i].set(tstr)
        for i in range(vv.ndim):
            ww = max(5, int(np.ceil(np.log10(vv.shape[i]))))  # 5~median
            self.vd[i].config(values=spinbox_values(vv.shape[i]), width=ww,
                              state=tk.NORMAL)
            if ( (dims[i] != latdim) and
                 (dims[i] != londim) and
                 (dims[i] != dunlim) and
                 (nall <= 1) and (vv.shape[i] > 1) ):
                nall += 1
                self.vdval[i].set('all')
                self.vdlblval[i].set(str(dims[i]))
                self.vdframe[i].pack(side='left')
                if vv.shape[i] > 1:
                    tstr  = "Specific dimension value: 0-{:d}\n".format(
                        vv.shape[i] - 1)
                    tstr += "or arithmetic operation on axis:\n"
                    tstr += "  " + ", ".join(DIMMETHODS)
                else:
                    tstr = "Single dimension: 0"
                self.vdtip[i].set(tstr)
            elif ((dims[i] != latdim) and
                  (dims[i] != londim)):
                self.vdval[i].set('0')
                self.vdlblval[i].set(str(dims[i]))
                self.vdframe[i].pack(side='left')
                if vv.shape[i] > 1:
                    tstr  = "Specific dimension value: 0-{:d}\n".format(
                        vv.shape[i] - 1)
                    tstr += "or arithmetic operation on axis:\n"
                    tstr += "  " + ", ".join(DIMMETHODS)
                else:
                    tstr = "Single dimension: 0"
                self.vdtip[i].set(tstr)


def set_dim_x(self):
    """
    Set spinboxes of x-dimensions.

    Set labels and value lists, including 'all' to select all entries,
    as well as 'mean', 'std', etc. for common operations on the axis.

    Select 'all' for the unlimited dimension if it exists, otherwise
    for the first dimension, and select 0 for all other dimensions.

    Parameters
    ----------
    self : class
        ncvue class

    Returns
    -------
    None
        Labels and values of spinboxes of x-dimensions set.

    Examples
    --------
    >>> set_dim_x(self)

    """
    # reset dimensions
    for i in range(self.maxdim):
        self.xd[i].config(values=(0,), width=1, state=tk.DISABLED)
        self.xdlblval[i].set(str(i))
        self.xdtip[i].set("")
        self.xdframe[i].pack(side='left')
    x = self.x.get()
    if x != '':
        # set real dimensions
        gx, vx = vardim2var(x, self.groups)
        if self.usex:
            if vx == self.tname:
                vx = self.tvar
        else:
            if vx == self.tname[gx]:
                vx = self.tvar[gx]
        xx = selvar(self, vx)
        if self.usex:
            dunlim = self.dunlim
            dims = xx.dims
        else:
            dunlim = self.dunlim[gx]
            dims = xx.dimensions
        nall = 0
        if dunlim in dims:
            i = dims.index(dunlim)
            ww = max(5, int(np.ceil(np.log10(xx.shape[i]))))  # 5~median
            self.xd[i].config(values=spinbox_values(xx.shape[i]), width=ww,
                              state=tk.NORMAL)
            nall += 1
            self.xdval[i].set('all')
            self.xdlblval[i].set(str(dims[i]))
            self.xdframe[i].pack(side='left')
            if xx.shape[i] > 1:
                tstr  = "Specific dimension value: 0-{:d}\n".format(
                    xx.shape[i] - 1)
                tstr += "or arithmetic operation on axis:\n"
                tstr += "  " + ", ".join(DIMMETHODS)
            else:
                tstr = "Single dimension: 0"
            self.xdtip[i].set(tstr)
        for i in range(xx.ndim):
            if dims[i] != dunlim:
                ww = max(5, int(np.ceil(np.log10(xx.shape[i]))))
                self.xd[i].config(values=spinbox_values(xx.shape[i]), width=ww,
                                  state=tk.NORMAL)
                if (nall == 0) and (xx.shape[i] > 1):
                    nall += 1
                    self.xdval[i].set('all')
                else:
                    self.xdval[i].set('0')
                self.xdlblval[i].set(str(dims[i]))
                self.xdframe[i].pack(side='left')
                if xx.shape[i] > 1:
                    tstr  = "Specific dimension value: 0-{:d}\n".format(
                        xx.shape[i] - 1)
                    tstr += "or arithmetic operation on axis:\n"
                    tstr += "  " + ", ".join(DIMMETHODS)
                else:
                    tstr = "Single dimension: 0"
                self.xdtip[i].set(tstr)


def set_dim_y(self):
    """
    Set spinboxes of y-dimensions of the left-hand-side (lhs).

    Set labels and value lists, including 'all' to select all entries,
    as well as 'mean', 'std', etc. for common operations on the axis.

    Select 'all' for the unlimited dimension if it exists, otherwise
    for the first dimension, and select 0 for all other dimensions.

    Parameters
    ----------
    self : class
        ncvue class

    Returns
    -------
    None
        Labels and values of spinboxes of lhs y-dimensions set.

    Examples
    --------
    >>> set_dim_y(self)

    """
    # reset dimensions
    for i in range(self.maxdim):
        self.yd[i].config(values=(0,), width=1, state=tk.DISABLED)
        self.ydlblval[i].set(str(i))
        self.ydtip[i].set("")
        self.ydframe[i].pack(side='left')
    y = self.y.get()
    if y != '':
        # set real dimensions
        gy, vy = vardim2var(y, self.groups)
        if self.usex:
            if vy == self.tname:
                vy = self.tvar
        else:
            if vy == self.tname[gy]:
                vy = self.tvar[gy]
        yy = selvar(self, vy)
        if self.usex:
            dunlim = self.dunlim
            dims = yy.dims
        else:
            dunlim = self.dunlim[gy]
            dims = yy.dimensions
        nall = 0
        if dunlim in dims:
            i = dims.index(dunlim)
            ww = max(5, int(np.ceil(np.log10(yy.shape[i]))))  # 5~median
            self.yd[i].config(values=spinbox_values(yy.shape[i]), width=ww,
                              state=tk.NORMAL)
            nall += 1
            self.ydval[i].set('all')
            self.ydlblval[i].set(str(dims[i]))
            self.ydframe[i].pack(side='left')
            if yy.shape[i] > 1:
                tstr  = "Specific dimension value: 0-{:d}\n".format(
                    yy.shape[i] - 1)
                tstr += "or arithmetic operation on axis:\n"
                tstr += "  " + ", ".join(DIMMETHODS)
            else:
                tstr = "Single dimension: 0"
            self.ydtip[i].set(tstr)
        for i in range(yy.ndim):
            if dims[i] != dunlim:
                ww = max(5, int(np.ceil(np.log10(yy.shape[i]))))
                self.yd[i].config(values=spinbox_values(yy.shape[i]), width=ww,
                                  state=tk.NORMAL)
                if (nall == 0) and (yy.shape[i] > 1):
                    nall += 1
                    self.ydval[i].set('all')
                else:
                    self.ydval[i].set('0')
                self.ydlblval[i].set(str(dims[i]))
                self.ydframe[i].pack(side='left')
                if yy.shape[i] > 1:
                    tstr  = "Specific dimension value: 0-{:d}\n".format(
                        yy.shape[i] - 1)
                    tstr += "or arithmetic operation on axis:\n"
                    tstr += "  " + ", ".join(DIMMETHODS)
                else:
                    tstr = "Single dimension: 0"
                self.ydtip[i].set(tstr)


def set_dim_y2(self):
    """
    Set spinboxes of y2-dimensions of the right-hand-side (rhs).

    Set labels and value lists, including 'all' to select all entries,
    as well as 'mean', 'std', etc. for common operations on the axis.

    Select 'all' for the unlimited dimension if it exists, otherwise
    for the first dimension, and select 0 for all other dimensions.

    Parameters
    ----------
    self : class
        ncvue class

    Returns
    -------
    None
        Labels and values of spinboxes of rhs y-dimensions set.

    Examples
    --------
    >>> set_dim_y2(self)

    """
    # reset dimensions
    for i in range(self.maxdim):
        self.y2d[i].config(values=(0,), width=1, state=tk.DISABLED)
        self.y2dlblval[i].set(str(i))
        self.y2dtip[i].set("")
        self.y2dframe[i].pack(side='left')
    y2 = self.y2.get()
    if y2 != '':
        # set real dimensions
        gy2, vy2 = vardim2var(y2, self.groups)
        if self.usex:
            if vy2 == self.tname:
                vy2 = self.tvar
        else:
            if vy2 == self.tname[gy2]:
                vy2 = self.tvar[gy2]
        yy2 = selvar(self, vy2)
        if self.usex:
            dunlim = self.dunlim
            dims = yy2.dims
        else:
            dunlim = self.dunlim[gy2]
            dims = yy2.dimensions
        nall = 0
        if dunlim in dims:
            i = dims.index(dunlim)
            ww = max(5, int(np.ceil(np.log10(yy2.shape[i]))))  # 5~median
            self.y2d[i].config(values=spinbox_values(yy2.shape[i]), width=ww,
                               state=tk.NORMAL)
            nall += 1
            self.y2dval[i].set('all')
            self.y2dlblval[i].set(str(dims[i]))
            self.y2dframe[i].pack(side='left')
            if yy2.shape[i] > 1:
                tstr  = "Specific dimension value: 0-{:d}\n".format(
                    yy2.shape[i] - 1)
                tstr += "or arithmetic operation on axis:\n"
                tstr += "  " + ", ".join(DIMMETHODS)
            else:
                tstr = "Single dimension: 0"
            self.y2dtip[i].set(tstr)
        for i in range(yy2.ndim):
            if dims[i] != dunlim:
                ww = max(5, int(np.ceil(np.log10(yy2.shape[i]))))
                self.y2d[i].config(values=spinbox_values(yy2.shape[i]),
                                   width=ww, state=tk.NORMAL)
                if (nall == 0) and (yy2.shape[i] > 1):
                    nall += 1
                    self.y2dval[i].set('all')
                else:
                    self.y2dval[i].set('0')
                self.y2dlblval[i].set(str(dims[i]))
                self.y2dframe[i].pack(side='left')
                if yy2.shape[i] > 1:
                    tstr  = "Specific dimension value: 0-{:d}\n".format(
                        yy2.shape[i] - 1)
                    tstr += "or arithmetic operation on axis:\n"
                    tstr += "  " + ", ".join(DIMMETHODS)
                else:
                    tstr = "Single dimension: 0"
                self.y2dtip[i].set(tstr)


def set_dim_z(self):
    """
    Set spinboxes of z-dimensions.

    Set labels and value lists, including 'all' to select all entries,
    as well as 'mean', 'std', etc. for common operations on the axis.

    Select 'all' for the unlimited dimension if it exists, otherwise
    for the first dimension, as well as for a second dimension and select
    0 for all other dimensions.

    Parameters
    ----------
    self : class
        ncvue class

    Returns
    -------
    None
        Labels and values of spinboxes of z-dimensions set.

    Examples
    --------
    >>> set_dim_z(self)

    """
    # reset dimensions
    for i in range(self.maxdim):
        self.zd[i].config(values=(0,), width=1, state=tk.DISABLED)
        self.zdlblval[i].set(str(i))
        self.zdtip[i].set("")
        self.zdframe[i].pack(side='left')
    z = self.z.get()
    if z != '':
        # set real dimensions
        gz, vz = vardim2var(z, self.groups)
        if self.usex:
            if vz == self.tname:
                vz = self.tvar
        else:
            if vz == self.tname[gz]:
                vz = self.tvar[gz]
        zz = selvar(self, vz)
        if self.usex:
            dunlim = self.dunlim
            dims = zz.dims
        else:
            dunlim = self.dunlim[gz]
            dims = zz.dimensions
        nall = 0
        if dunlim in dims:
            i = dims.index(dunlim)
            ww = max(5, int(np.ceil(np.log10(zz.shape[i]))))  # 5~median
            self.zd[i].config(values=spinbox_values(zz.shape[i]), width=ww,
                              state=tk.NORMAL)
            nall += 1
            self.zdval[i].set('all')
            self.zdlblval[i].set(str(dims[i]))
            self.zdframe[i].pack(side='left')
            if zz.shape[i] > 1:
                tstr  = "Specific dimension value: 0-{:d}\n".format(
                    zz.shape[i] - 1)
                tstr += "or arithmetic operation on axis:\n"
                tstr += "  " + ", ".join(DIMMETHODS)
            else:
                tstr = "Single dimension: 0"
            self.zdtip[i].set(tstr)
        for i in range(zz.ndim):
            if dims[i] != dunlim:
                ww = max(5, int(np.ceil(np.log10(zz.shape[i]))))
                self.zd[i].config(values=spinbox_values(zz.shape[i]), width=ww,
                                  state=tk.NORMAL)
                if (nall <= 1) and (zz.shape[i] > 1):
                    nall += 1
                    self.zdval[i].set('all')
                else:
                    self.zdval[i].set('0')
                self.zdlblval[i].set(str(dims[i]))
                self.zdframe[i].pack(side='left')
                if zz.shape[i] > 1:
                    tstr  = "Specific dimension value: 0-{:d}\n".format(
                        zz.shape[i] - 1)
                    tstr += "or arithmetic operation on axis:\n"
                    tstr += "  " + ", ".join(DIMMETHODS)
                else:
                    tstr = "Single dimension: 0"
                self.zdtip[i].set(tstr)



#!/usr/bin/env python
"""
usage: ncvue [-h] [-m missing_value] [netcdf_file]

A minimal GUI for a quick view of netcdf files.

positional arguments:
  netcdf_file           netcdf file

optional arguments:
  -h, --help            show this help message and exit
  -m missing_value, --miss missing_value
                        Set value to missing value (default: numpy.nan)
  -x, --xarray          Use xarray to read input files


Example command line:
    ncvue MuSICA_out_2009.nc


:copyright: Copyright 2020- Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

History
   * Written Nov-Dec 2020 by Matthias Cuntz (mc (at) macu (dot) de)
   * Made ncvue a gui_script entry_point, May 2022, Matthias Cuntz
   * Include xarray to read input files, Feb 2025, Matthias Cuntz

"""
import numpy as np
from ncvue import ncvue


def main():
    import argparse

    miss = np.nan
    usex = False
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description='A minimal GUI for a quick view of netcdf files.')
    hstr = 'Set value to missing value (default: numpy.nan)'
    parser.add_argument('-m', '--miss', action='store', type=float,
                        default=miss, dest='miss',
                        metavar='missing_value', help=hstr)
    hstr = 'Use xarray to read input files'
    parser.add_argument('-x', '--xarray', action='store_true',
                        default=usex, dest='usex', help=hstr)
    parser.add_argument('ncfile', nargs='*', default=None,
                        metavar='netcdf_file',
                        help='netcdf file')

    args = parser.parse_args()
    miss = args.miss
    usex = args.usex
    ncfile = args.ncfile

    del parser, args

    # if len(ncfile) == 0:
    #     ncfile = ncfile[0]
    # else:
    #     ncfile = ''

    # This must be before any other call to matplotlib
    # because it uses the TkAgg backend.
    # This means, do not use --pylab with ipython.
    ncvue(ncfile=ncfile, miss=miss, usex=usex)


if __name__ == "__main__":
    main()
